package biggests.clustering;

import java.io.*;
import java.text.*;
import java.util.*;

import smadeira.biclustering.*;
import smadeira.utils.TaskPercentages;

/**
 * <p>Title: Hierarchical Clustering</p>
 *
 * <p>Description: This class implements Hierarchical Clustering over
 *                 expression data. Most of the methods are based in
 *                 the C source code of Cluster 3.0 (version 1.36) by
 *                 Michiel de Hoon, while at the Laboratory of DNA
 *                 Information Analysis, Human Genome Center, Institute
 *                 of Medical Science, University of Tokyo, Japan.
 * <br/>
 *                 Cluster 3.0 is an enhanced version of Cluster,
 *                 which was originally developed by Michael Eisen
 *                 while at Stanford University.
 * </p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 2 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 *
 *                 GNU General Public License also complies with the terms of
 *                 Cluster/TreeView
 *                 <a href="http://rana.lbl.gov/EisenSoftwareSource.htm">
 *                 original license</a>
 * </p>
 *
 * @author Joana P. Gonalves
 */
public class HierarchicalClustering extends AbstractMetricClustering {
    /**
     * Pairwise complete-linkage (clustering method).
     */
    public static final char COMPLETE_LINKAGE = 'm';

    /**
     * Pairwise single-linkage (clustering method).
     */
    public static final char SINGLE_LINKAGE = 's';

    /**
     * Pairwise centroid-linkage (clustering method).
     */
    public static final char CENTROID_LINKAGE = 'c';

    /**
     * Pairwise average linkage (clustering method).
     */
    public static final char AVERAGE_LINKAGE = 'a';

    /**
     * Creates a new <code>HierarchicalClustering</code> object to
     * perform hierarchical clustering over the set of gene expression
     * data contained in <code>expressionMatrix</code>.
     *
     * @param expressionMatrix <code>IMatrix</code>
     */
    public HierarchicalClustering(IMatrix expressionMatrix) {
        super(expressionMatrix);
    }

    /**
     * Creates a new <code>HierarchicalClustering</code> object to
     * perform hierarchical clustering over the given set of
     * expression data.
     *
     * @param numberOfGenes <code>int</code> the number of genes in the
     *        experiment
     * @param numberOfConditions <code>int</code> the number of conditions
     *        in the experiment
     * @param genesWeights <code>double[]</code> the set of genes' weights
     * @param conditionsWeights <code>double[]</code> the set of conditions'
     *        weights
     * @param genesOrder <code>double[]</code> the order of the genes
     * @param conditionsOrder <code>double[]</code> the order of the conditions
     * @param genesIndexes <code>int[]</code> the set of genes' indexes
     * @param conditionsIndexes <code>int[]</code> the set of conditions'
     *        indexes
     * @param uniqueID <code>String</code> the unique ID type name
     * @param genesUniqueIDs <code>String[]</code> the set of genes' unique IDs
     * @param genesNames <code>String[]</code> the set of genes' names
     * @param conditionsNames <code>String[]</code> the set of conditions' names
     * @param expressionData <code>double[][]</code> the set of expression
     *        values measured for the genes in the experimental conditions
     * @param missingsMask <code>int[][]</code> the mask which states if each
     *        element contained in <code>expressionData</code> is valid (1) or
     *        missing (0)
     * @throws Exception - if array and/or matrices dimensions do not match
     *        the <code>numberOfGenes</code> and/or
     *        <code>numberOfConditions</code>
     */
    public HierarchicalClustering(int numberOfGenes,
                                  int numberOfConditions,
                                  double[] genesWeights,
                                  double[] conditionsWeights,
                                  double[] genesOrder,
                                  double[] conditionsOrder,
                                  int[] genesIndexes,
                                  int[] conditionsIndexes,
                                  String uniqueID,
                                  String[] genesUniqueIDs,
                                  String[] genesNames,
                                  String[] conditionsNames,
                                  double[][] expressionData,
                                  int[][] missingsMask) throws Exception {
        super(numberOfGenes,
              numberOfConditions,
              genesWeights,
              conditionsWeights,
              genesOrder,
              conditionsOrder,
              genesIndexes,
              conditionsIndexes,
              uniqueID,
              genesUniqueIDs,
              genesNames,
              conditionsNames,
              expressionData,
              missingsMask);
    }


    /**
     * Performs hierarchical clustering over this set of data and writes
     * output information in corresponding files (.GTR, .ATR, .CDT).
     *
     * @param genesMetric <code>char</code> the distance metric to use in
     *        genes' clustering; available distance metrics are:<br/>
     *        <ul>
     *        <li><code>NO_CLUSTERING</code> -
     *        with this option no clustering will be performed on the
     *        genes' dimension
     *        <li><code>UNCENTERED_CORRELATION</code>
     *        <li><code>PEARSON_CORRELATION</code>
     *        <li><code>UNCENTERED_ABS_CORRELATION</code>
     *        <li><code>PEARSON_ABS_CORRELATION</code>
     *        <li><code>SPEARMAN_CORRELATION</code>
     *        <li><code>KENDALL_TAU_CORRELATION</code>
     *        <li><code>EUCLIDEAN_DISTANCE</code>
     *        <li><code>CITYBLOCK_DISTANCE</code>
     *        </ul>
     * @param conditionsMetric <code>char</code> the distance metric to use
     *        in conditions' clustering; available distance metrics are:<br/>
     *        <ul>
     *        <li><code>NO_CLUSTERING</code> -
     *        with this option no clustering will be performed on the
     *        columns' dimension
     *        <li><code>UNCENTERED_CORRELATION</code>
     *        <li><code>PEARSON_CORRELATION</code>
     *        <li><code>UNCENTERED_ABS_CORRELATION</code>
     *        <li><code>PEARSON_ABS_CORRELATION</code>
     *        <li><code>SPEARMAN_CORRELATION</code>
     *        <li><code>KENDALL_TAU_CORRELATION</code>
     *        <li><code>EUCLIDEAN_DISTANCE</code>
     *        <li><code>CITYBLOCK_DISTANCE</code>
     *        </ul>
     * @param clusteringMethod <code>char</code> the clustering method to apply;
     *        available clustering methods are:
     *        <ul>
     *        <li><code>COMPLETE_LINKAGE</code>
     *        <li><code>SINGLE_LINKAGE</code>
     *        <li><code>AVERAGE_LINKAGE</code>
     *        <li><code>CENTROID_LINKAGE</code>
     *        </ul>
     * @param jobName <code>String</code> the name to appear in all
     *        output filenames, before the file extension
     * @throws IOException - if files could not be opened for writing or
     * some problem occured while writing to them
     */
    public void hierarchicalClusteringWithFileOutput(char genesMetric,
            char conditionsMetric,
            char clusteringMethod,
            String jobName) throws IOException {
        String fileName;
        if (genesMetric != NO_CLUSTERING) {
            progressBar.setString("clustering genes...");
            setTaskPercentages(new TaskPercentages(
                    TaskPercentages.HIERARCHICAL_CLUSTERING));
            fileName = jobName + ".gtr";
            hierarchicalClusteringWithFileOutputSingleDimension(fileName,
                    genesMetric, false, clusteringMethod);
        }
        if (conditionsMetric != NO_CLUSTERING) {
            progressBar.setString("clustering conditions...");
            setTaskPercentages(new TaskPercentages(
                    TaskPercentages.HIERARCHICAL_CLUSTERING));
            fileName = jobName + ".atr";
            hierarchicalClusteringWithFileOutputSingleDimension(fileName,
                    conditionsMetric, true, clusteringMethod);
        }
        fileName = jobName + ".cdt";
        writeCDTFile(fileName, genesMetric != NO_CLUSTERING,
                     conditionsMetric != NO_CLUSTERING);
    }

    /**
     * Performs hierarchical clustering over the genes' or conditions'
     * dimension and writes the results to the corresponding .GRT or .ATR
     * file with the name indicated by <code>fileName</code>.
     *
     * @param fileName <code>String</code> the name of the file to write
     *        the output of the hierarchical clustering (it contains the
     *        correct file extension - .gtr for genes' clustering, .atr
     *        for conditions' clustering)
     * @param clusteringMetric <code>char</code> the clustering metric to
     *        use in clustering algorithm; available clustering metrics are:<br/>
     *        <ul>
     *        <li><code>NO_CLUSTERING</code> -
     *        this option should not be used with this method
     *        <li><code>UNCENTERED_CORRELATION</code>
     *        <li><code>PEARSON_CORRELATION</code>
     *        <li><code>UNCENTERED_ABS_CORRELATION</code>
     *        <li><code>PEARSON_ABS_CORRELATION</code>
     *        <li><code>SPEARMAN_CORRELATION</code>
     *        <li><code>KENDALL_TAU_CORRELATION</code>
     *        <li><code>EUCLIDEAN_DISTANCE</code>
     *        <li><code>CITYBLOCK_DISTANCE</code>
     *        </ul>
     * @param clusteringDimension <code>boolean</code> states in which
     *        dimension clustering should be applied:
     *        <ul>
     *        <li><code>GENES_DIMENSION</code>
     *        <li><code>CONDITIONS_DIMENSION</code>
     *        </ul>
     * @param clusteringMethod <code>char</code> the clustering
     *        algorithm/method to use:
     *        <ul>
     *        <li><code>COMPLETE_LINKAGE</code>
     *        <li><code>SINGLE_LINKAGE</code>
     *        <li><code>AVERAGE_LINKAGE</code>
     *        <li><code>CENTROID_LINKAGE</code>
     *        </ul>
     * @return boolean indicates the success of the clustering operation;
     *         <code>true</code> for success, <code>false</code> otherwise
     * @throws ArrayIndexOutOfBoundsException
     * @throws FileNotFoundException if the file specified by
     *         <code>fileName</code> could not be opened for writing
     * @throws IOException if some problem occured while writing results
     *         to file
     * @throws Exception
     */
    private boolean hierarchicalClusteringWithFileOutputSingleDimension(
            String fileName,
            char clusteringMetric,
            boolean clusteringDimension,
            char clusteringMethod) throws FileNotFoundException, IOException {
        int i;
        double[] nodesOrder;
        int[] nodesCounts;
        String[] nodesIDs;

        int numberOfNodes =
                (clusteringDimension == this.CONDITIONS_DIMENSION ?
                 numberOfConditions : numberOfGenes) - 1;
        double[] clusteringDimensionOrder =
                (clusteringDimension == this.GENES_DIMENSION) ?
                genesOrder : conditionsOrder;
        double[] elementsWeights =
                (clusteringDimension == this.GENES_DIMENSION) ?
                conditionsWeights : genesWeights;
        String keyword = (clusteringDimension == this.GENES_DIMENSION) ? "GENE" :
                         "ARRY";

        /* Perform hierarchical clustering. */
        ClusterTreeNode[] tree = null;
        try {
            tree = hierarchicalClusteringWithTreeOutput(expressionData,
                    missingsMask,
                    elementsWeights,
                    clusteringDimension,
                    clusteringMetric,
                    clusteringMethod, null);
        } catch (Exception ex) {
            ex.printStackTrace();
            /* computeDistance throws an ArrayIndexOutOfBoundsException
             or an Exception if the distanceMatrix is not valid, but since
             we are going to compute it with computeDistanceMatrix and given
                         valid data, we won't have to worry about this */
        }
        if (tree == null) {
            return false;
        }

        if (clusteringMetric == EUCLIDEAN_DISTANCE ||
            clusteringMetric == CITYBLOCK_DISTANCE) {
            /* Scale all distances such that they are between 0 and 1 */
            double scale = 0.0;
            for (i = 0; i < numberOfNodes; i++) {
                if (tree[i].getDistance() > scale) {
                    scale = tree[i].getDistance();
                }
            }
            if (scale != 0) {
                for (i = 0; i < numberOfNodes; i++) {
                    tree[i].setDistance(tree[i].getDistance() * 1.0 / scale);
                }
            }
        }

        /* Now we join nodes */
        nodesOrder = new double[numberOfNodes];
        nodesCounts = new int[numberOfNodes];
        nodesIDs = new String[numberOfNodes];

        PrintWriter pw = new PrintWriter(fileName);
        NumberFormat nf = NumberFormat.getInstance(Locale.ENGLISH);
        nf.setMinimumFractionDigits(6);
        nf.setMaximumFractionDigits(6);
        for (i = 0; i < numberOfNodes; i++) {
            int min1 = tree[i].getLeft();
            int min2 = tree[i].getRight();
            /* min1 and min2 are the elements that are to be joined */
            double order1;
            double order2;
            int counts1;
            int counts2;
            String ID1;
            String ID2;
            nodesIDs[i] = "NODE" + (i + 1) + "X";
            if (min1 < 0) {
                int index1 = -min1 - 1;
                order1 = nodesOrder[index1];
                counts1 = nodesCounts[index1];
                ID1 = nodesIDs[index1];
                tree[i].setDistance(Math.max(tree[i].getDistance(),
                                             tree[index1].getDistance()));
            } else {
                order1 = clusteringDimensionOrder[min1];
                counts1 = 1;
                ID1 = keyword + min1 + "X";
            }
            if (min2 < 0) {
                int index2 = -min2 - 1;
                order2 = nodesOrder[index2];
                counts2 = nodesCounts[index2];
                ID2 = nodesIDs[index2];
                tree[i].setDistance(Math.max(tree[i].getDistance(),
                                             tree[index2].getDistance()));
            } else {
                order2 = clusteringDimensionOrder[min2];
                counts2 = 1;
                ID2 = keyword + min2 + "X";
            }

            pw.print(nodesIDs[i] + "\t" + ID1 + "\t" + ID2 + "\t");
            pw.print(nf.format(1.0 - tree[i].getDistance()) + "\n");

            nodesCounts[i] = counts1 + counts2;
            nodesOrder[i] = (counts1 * order1 + counts2 * order2) * 1.0 /
                            (counts1 + counts2);
        }
        pw.close();

        /* Now set up order based on the tree structure */
        try {
            sortClusteringTree(clusteringDimension,
                               clusteringDimensionOrder,
                               nodesOrder,
                               nodesCounts,
                               tree);
        } catch (Exception ex) {
            ex.printStackTrace();
            /* */
        }
        return true;
    }

    /**
     * Writes CDT file with genes' names and unique IDs, conditions' names,
     * genes' weights and conditions' weights accordingo to genes' indexes
     * and conditions' indexes. Also writes expression data values
     * according to the missing values' mask. The output file will have
     * the following format:<br/>
     * <table>
     * <tr><th>GID<th>UNIQID<th>NAME<th>GWEIGHT<td>C1name<td>C2name<td>C3name
     * <tr><th>EWEIGHT<td><td><td><td>C1weight<td>C2weight<td>C3weight
     * <tr><td>G1id<td>G1uniqID<td>G1name<td>G1weight<td>value11<td>value12<td>value13
     * <tr><td>G2id<td>G2uniqID<td>G2name<td>G2weight<td>value21<td>value22<td>value23
     * <tr><td>G3id<td>G3uniqID<td>G3name<td>G3weight<td>value31<td>value32<td>value33
     * <tr><td>G4id<td>G4uniqID<td>G4name<td>G4weight<td>value41<td>value42<td>value43
     * </table>
     *
     * @param fileName <code>String</code> the name of the output CDT file
     *        (with the .cdt file extension)
     * @param withGenesIDs <code>boolean</code> <code>true</code> if genes
     *        were clustered, <code>false</code> otherwise
     * @param withConditionsIDs <code>boolean</code> <code>true</code> if
     *        conditions were clustered, <code>false</code> otherwise
     * @throws FileNotFoundException - if the file with the name
     *         <code>fileName</code> couldn't be opened for writing
     * @throws IOException - if some problem occured while writing to file
     */
    private void writeCDTFile(String fileName,
                              boolean withGenesIDs,
                              boolean withConditionsIDs) throws
            FileNotFoundException, IOException {
        PrintWriter pw = new PrintWriter(fileName);
        if (withGenesIDs) {
            pw.print("GID\t");
        }
        pw.print(uniqueID);
        pw.print("\tNAME\tGWEIGHT");
        /* Now add headers for data columns */
        for (int column = 0; column < numberOfConditions; column++) {
            pw.print("\t");
            pw.print(conditionsNames[conditionsIndexes[column]]);
        }
        pw.print("\n");

        if (withConditionsIDs) {
            pw.print("AID");
            if (withGenesIDs) {
                pw.print("\t");
            }
            pw.print("\t\t");
            for (int column = 0; column < numberOfConditions; column++) {
                String ID = "ARRY" + conditionsIndexes[column] + "X";
                pw.print("\t");
                pw.print(ID);
            }
            pw.print("\n");
        }

        pw.print("EWEIGHT");
        if (withGenesIDs) {
            pw.print("\t");
        }
        pw.print("\t\t");
        for (int column = 0; column < numberOfConditions; column++) {
            pw.print("\t" + conditionsWeights[conditionsIndexes[column]]);
        }
        pw.print("\n");

        for (int row = 0; row < numberOfGenes; row++) {
            int index = genesIndexes[row];
            if (withGenesIDs) {
                String ID = "GENE" + index + "X";
                pw.print(ID);
                pw.print("\t");
            }

            pw.print(genesUniqueIDs[index]);
            pw.print("\t");
            if (genesNames[index] != null) {
                pw.print(genesNames[index]);
            } else {
                pw.print(genesUniqueIDs[index]);
            }
            pw.print("\t" + genesWeights[index]);

            NumberFormat nf = NumberFormat.getInstance(Locale.ENGLISH);
            nf.setMinimumFractionDigits(6);
            nf.setMaximumFractionDigits(6);

            for (int column = 0; column < numberOfConditions; column++) {
                int columnindex = conditionsIndexes[column];
                pw.print("\t");
                if (missingsMask[index][columnindex] == 1) {
                    pw.print(nf.format(expressionData[index][columnindex]));
                }
            }
            pw.print("\n");
        }
        pw.close();
        return;
    }

    /**
     * Sorts the nodes of the <code>tree</code> based on the
     * tree's structure.
     *
     * @param sortingDimension <code>boolean</code> states in which
     *        dimension clustering was performed, which is also the
     *        same dimension to perform the sorting operation; the
     *        available dimensions are:
     *        <ul>
     *        <li><code>GENES_DIMENSION</code>
     *        <li><code>CONDITIONS_DIMENSION</code>
     *        </ul>
     * @param clusteringDimensionOrder <code>double[]</code> the order
     *        of the genes or conditions, according to the dimension in
     *        which the tree will be sorted
     * @param nodesOrder <code>double[]</code> the order of the nodes
     * @param nodesCounts <code>int[]</code> the number of elements for
     *        each node
     * @param tree <code>Node[]</code> the set of nodes that make up the tree
     * @throws Exception
     */
    public void sortClusteringTree(boolean sortingDimension,
                                   double[] clusteringDimensionOrder,
                                   double[] nodesOrder,
                                   int[] nodesCounts,
                                   ClusterTreeNode[] tree) throws Exception {
        if (sortingDimension == GENES_DIMENSION &&
            (nodesOrder.length != genesIndexes.length - 1 ||
             nodesCounts.length != genesIndexes.length - 1 ||
             tree.length != genesIndexes.length - 1)) {
            throw new Exception("nodesOrder, nodesCounts and tree length " +
                                "must be equal to the length of genesIndexes-1");
        }
        if (sortingDimension == CONDITIONS_DIMENSION &&
            (nodesOrder.length != conditionsIndexes.length - 1 ||
             nodesCounts.length != conditionsIndexes.length - 1 ||
             tree.length != conditionsIndexes.length - 1)) {
            throw new Exception("nodesOrder, nodesCounts and tree length " +
                                "must be equal to the length of conditionsIndexes-1");
        }
        double[] newOrder = new double[tree.length + 1];
        for (int i = 0; i < newOrder.length; i++) {
            newOrder[i] = 0.0;
        }
        int[] clusterIDs = new int[tree.length + 1];
        for (int i = 0; i < tree.length + 1; i++) {
            clusterIDs[i] = i;
        }

        for (int i = 0; i < tree.length; i++) {
            int i1 = tree[i].getLeft();
            int i2 = tree[i].getRight();
            double order1 = (i1 < 0) ? nodesOrder[ -i1 - 1] :
                            clusteringDimensionOrder[i1];
            double order2 = (i2 < 0) ? nodesOrder[ -i2 - 1] :
                            clusteringDimensionOrder[i2];
            int count1 = (i1 < 0) ? nodesCounts[ -i1 - 1] : 1;
            int count2 = (i2 < 0) ? nodesCounts[ -i2 - 1] : 1;
            /* If order1 and order2 are equal, their order is determined by
             * the order in which they were clustered */
            if (i1 < i2) {
                double increase = (order1 < order2) ? count1 : count2;
                int j;
                for (j = 0; j < tree.length + 1; j++) {
                    int clusterID = clusterIDs[j];
                    if (clusterID == i1 && order1 >= order2) {
                        newOrder[j] += increase;
                    }
                    if (clusterID == i2 && order1 < order2) {
                        newOrder[j] += increase;
                    }
                    if (clusterID == i1 || clusterID == i2) {
                        clusterIDs[j] = -i - 1;
                    }
                }
            } else {
                double increase = (order1 <= order2) ? count1 : count2;
                int j;
                for (j = 0; j < tree.length + 1; j++) {
                    int clusterID = clusterIDs[j];
                    if (clusterID == i1 && order1 > order2) {
                        newOrder[j] += increase;
                    }
                    if (clusterID == i2 && order1 <= order2) {
                        newOrder[j] += increase;
                    }
                    if (clusterID == i1 || clusterID == i2) {
                        clusterIDs[j] = -i - 1;
                    }
                }
            }
        }
        if (sortingDimension == GENES_DIMENSION) {
            for (int i = 0; i < numberOfGenes; i++) {
                genesIndexes[i] = i;
            }
            sortIndexes(newOrder, genesIndexes);
        }
        if (sortingDimension == CONDITIONS_DIMENSION) {
            for (int i = 0; i < numberOfConditions; i++) {
                conditionsIndexes[i] = i;
            }
            sortIndexes(newOrder, conditionsIndexes);
        }
        return;
    }

    /**
     * Performs hierarchical clustering using pairwise single-, maximum-,
     * centroid-, or average-linkage, as defined by
     * <code>clusteringMethod</code>, on a given set of gene expression
     * data, using the distance metric given by <code>distanceMetric</code>.
     * If successful, the function returns a pointer to a newly allocated
     * array of <code>Node</code>s containing the hierarchical clustering
     * solution.
     *
     * @param expressionData <code>double[][]</code> the data that contains
     *        the single dimension expression arrays (rows/genes or
     *        columns/conditions) to be clustered; first dimension contains
     *        genes/rows, second dimension contains conditions/columns
     * @param missingsMask <code>int[][]</code> shows which expression
     *        values are missing; if <code>missingsMask[i][j]</code> == 0,
     *        then <code>expressionData[i][j]</code> is missing; otherwise,
     *        if <code>missingsMask[i][j]</code> == 1, then
     *        <code>expressionData[i][j]</code> is valid
     * @param elementsWeights <code>double[]</code> the weights
     *        of the elements to be clustered (genes/rows or
     *        conditions/columns);  the length of this array is equal to
     *        the number of conditions/columns if the distances between
     *        genes/rows are calculated, or the number of genes/rows if the
     *        distances between conditions/columns are calculated
     * @param clusteringDimension </code>boolean</code> the clustering
     *        dimension; available dimensions:
     *        <ul>
     *        <li><code>GENES_DIMENSION</code> - the genes/rows of
     *        <code>expressionData</code> are clustered
     *        <li><code>CONDITIONS_DIMENSION</code> - the conditions/columns
     *        of <code>expressionData</code> are clustered
     *        </ul>
     * @param distanceMetric <code>char</code> the distance measure to use;
     *        valid distance metrics are:<br/>
     *        <ul>
     *        <li><code>UNCENTERED_CORRELATION</code>
     *        <li><code>PEARSON_CORRELATION</code>
     *        <li><code>UNCENTERED_ABS_CORRELATION</code>
     *        <li><code>PEARSON_ABS_CORRELATION</code>
     *        <li><code>SPEARMAN_CORRELATION</code>
     *        <li><code>KENDALL_TAU_CORRELATION</code>
     *        <li><code>EUCLIDEAN_DISTANCE</code>
     *        <li><code>CITYBLOCK_DISTANCE</code>
     *        for any other value, EUCLIDEAN_DISTANCE will be used
     *        </ul>
     * @param clusteringMethod <code>char</code> the clustering
     *        algorithm/method to use:
     *        <ul>
     *        <li><code>COMPLETE_LINKAGE</code>
     *        <li><code>SINGLE_LINKAGE</code>
     *        <li><code>AVERAGE_LINKAGE</code>
     *        <li><code>CENTROID_LINKAGE</code>
     *        </ul>
     *        for the first three, either the distance matrix or the gene
     *        expression data is sufficient to perform the clustering
     *        algorithm; for pairwise centroid-linkage clustering, however,
     *        the gene expression data are always needed, even if the
     *        distance matrix itself is available
     * @param distanceMatrix <code>double[][]</code> the distance matrix;
     *        if the distance matrix is null initially, it will be
     *        calculated from the data; if the given distance matrix is
     *        not null, its contents will be modified as part of the
     *        clustering algorithm
     * @return Node[] the set of <code>Node</code> objects, describing the
     *        hierarchical clustering solution consisting of
     *        (clusteredElements-1) nodes; depending on whether genes (rows)
     *        or conditions (columns) were clustered, clusteredElements is
     *        equal to the number of genes/rows or the number of
     *        conditions/columns, respectively
     * @throws ArrayIndexOutOfBoundsException
     * @throws Exception
     */
    public ClusterTreeNode[] hierarchicalClusteringWithTreeOutput(
            double[][] expressionData,
            int[][] missingsMask,
            double[] elementsWeights,
            boolean clusteringDimension,
            char distanceMetric,
            char clusteringMethod,
            double[][] distanceMatrix) throws ArrayIndexOutOfBoundsException,
            Exception {
        int numberOfElements = (clusteringDimension == GENES_DIMENSION) ?
                               expressionData.length : expressionData[0].length;

        if (numberOfElements < 2) {
            return null;
        }

        /* Calculate the distance matrix if the user didn't give it */
        if (distanceMatrix == null && clusteringMethod != SINGLE_LINKAGE) {
            distanceMatrix =
            computeDistanceMatrix(expressionData, missingsMask,
                                  elementsWeights,
                                  distanceMetric, clusteringDimension);
            this.incrementPercentDone();
        }

        switch (clusteringMethod) {
        case SINGLE_LINKAGE:
            return singleLinkageClustering(expressionData, missingsMask,
                                           elementsWeights, distanceMatrix,
                                           distanceMetric, clusteringDimension);
        case COMPLETE_LINKAGE:
            return completeLinkageClustering(distanceMatrix);
        case AVERAGE_LINKAGE:
            return averageLinkageClustering(distanceMatrix);
        case CENTROID_LINKAGE:
            return centroidLinkageClustering(expressionData, missingsMask,
                                             elementsWeights, distanceMatrix,
                                             distanceMetric,
                                             clusteringDimension);
        }
        return null;
    }

    /**
     * Calculates the distance matrix between genes/rows or
     * conditions/columns using their measured gene expression data.
     * Several distance measures can be used. Returns a
     * ragged array containing the distances between the elements
     * (genes/rows or conditions/columns). As the distance matrix is
     * symmetric, with zeros on the diagonal, only the lower triangular
     * half of the distance matrix is saved.
     *
     * @param expressionData <code>double[][]</code> the data that contains
     *        the single dimension expression arrays (rows/genes or
     *        columns/conditions) to be clustered; first dimension contains
     *        genes/rows, second dimension contains conditions/columns
     * @param missingsMask <code>int[][]</code> shows which expression
     *        values are missing; if <code>missingsMask[i][j]</code> == 0,
     *        then <code>expressionData[i][j]</code> is missing; otherwise,
     *        if <code>missingsMask[i][j]</code> == 1, then
     *        <code>expressionData[i][j]</code> is valid
     * @param elementsWeights <code>double[]</code> the weights
     *        of the elements to be clustered (genes/rows or
     *        conditions/columns); the length of this array is equal to
     *        the number of conditions/columns if the distances between
     *        genes/rows are calculated, or the number of genes/rows if the
     *        distances between conditions/columns are calculated
     * @param distanceMetric <code>char</code> the distance measure to use;
     *        valid distance metrics are:<br/>
     *        <ul>
     *        <li><code>UNCENTERED_CORRELATION</code>
     *        <li><code>PEARSON_CORRELATION</code>
     *        <li><code>UNCENTERED_ABS_CORRELATION</code>
     *        <li><code>PEARSON_ABS_CORRELATION</code>
     *        <li><code>SPEARMAN_CORRELATION</code>
     *        <li><code>KENDALL_TAU_CORRELATION</code>
     *        <li><code>EUCLIDEAN_DISTANCE</code>
     *        <li><code>CITYBLOCK_DISTANCE</code>
     *        for any other value, EUCLIDEAN_DISTANCE will be used
     *        </ul>
     * @param clusteringDimension </code>boolean</code> the clustering
     *        dimension; available dimensions:
     *        <ul>
     *        <li><code>GENES_DIMENSION</code> - the genes/rows of
     *        <code>expressionData</code> are clustered
     *        <li><code>CONDITIONS_DIMENSION</code> - the conditions/columns
     *        of <code>expressionData</code> are clustered
     *        </ul>
     * @return double[][] the computed distance matrix
     * @throws Exception
     */
    public static double[][] computeDistanceMatrix(double[][] expressionData,
            int[][] missingsMask,
            double[] elementsWeights,
            char distanceMetric,
            boolean clusteringDimension) throws Exception {
        if (expressionData.length != missingsMask.length ||
            expressionData[0].length != missingsMask[0].length) {
            throw new Exception("expressionData and missingsMask must " +
                                "have the equal dimensions");
        }
        /* First determine the size of the distance matrix */
        int numberOfElements = (clusteringDimension == GENES_DIMENSION) ?
                               expressionData.length : expressionData[0].length;
        /*        if (elementsWeights.length != numberOfElements) {
         throw new Exception ("elementsWeights must have the same " +
                                          "length than the corresponding " +
         "expressionData dimension (genes' or " +
                                          "conditions') that is being " +
                                          "taken for measurement");

                }
         */
        double[][] distanceMatrix;

        if (numberOfElements < 2) {
            return null;
        }

        /* Set up the ragged array */
        distanceMatrix = new double[numberOfElements][];
        distanceMatrix[0] = null;

        /* The zeroth row has zero columns. We allocate it anyway for convenience.*/
        for (int i = 1; i < numberOfElements; i++) {
            distanceMatrix[i] = new double[i];
        }

        /* Calculate the distances and save them in the ragged array */
        for (int i = 1; i < numberOfElements; i++) {
            for (int j = 0; j < i; j++) {
                distanceMatrix[i][j] = computeDistance(expressionData,
                        expressionData,
                        missingsMask,
                        missingsMask,
                        elementsWeights,
                        i,
                        j,
                        clusteringDimension,
                        distanceMetric);
            }
        }

        return distanceMatrix;
    }

    /**
     * Performs single-linkage hierarchical clustering, using either
     * the distance matrix directly, if available, or by calculating the
     * distances from the expression data matrix. This implementation is
     * based on the SLINK algorithm, described in:
     * Sibson, R. (1973). SLINK: <i>An optimally efficient algorithm for
     * the single-link cluster method.</i> The Computer Journal, 16(1): 30-34.
     * The output of this algorithm is identical to conventional
     * single-linkage hierarchical clustering, but is much more
     * memory-efficient and faster. Hence, it can be applied to large
     * data sets, for which the conventional single-linkage algorithm
     * fails due to lack of memory.
     *
     * @param expressionData <code>double[][]</code> the data that contains
     *        the single dimension expression arrays (rows/genes or
     *        columns/conditions) to be clustered; first dimension contains
     *        genes/rows, second dimension contains conditions/columns
     * @param missingsMask <code>int[][]</code> shows which expression
     *        values are missing; if <code>missingsMask[i][j]</code> == 0,
     *        then <code>expressionData[i][j]</code> is missing; otherwise,
     *        if <code>missingsMask[i][j]</code> == 1, then
     *        <code>expressionData[i][j]</code> is valid
     * @param elementsWeights <code>double[]</code> the weights
     *        of the elements to be clustered (genes/rows or
     *        conditions/columns); the length of this array is equal to
     *        the number of conditions/columns if the distances between
     *        genes/rows are calculated, or the number of genes/rows if the
     *        distances between conditions/columns are calculated
     * @param distanceMatrix double[][] the distance matrix; if the
     *        distance matrix is not null, it is used to speed up the
     *        clustering calculation and the original expression data
     *        (specified by <code>expressionData</code> and
     *        <code>missingsMask</code>) are not needed and, therefore,
     *        ignored; the contents of the distance matrix
     *        are not modified; if <code>distanceMatrix</code> is null,
     *        the pairwise distances are calculated from the gene expression
     *        data (the <code>expressionData</code> and
     *        <code>missingsMask</code> bidimensional arrays) and stored in
     *        temporary arrays
     * @param distanceMetric <code>char</code> the distance measure to use;
     *        valid distance metrics are:<br/>
     *        <ul>
     *        <li><code>UNCENTERED_CORRELATION</code>
     *        <li><code>PEARSON_CORRELATION</code>
     *        <li><code>UNCENTERED_ABS_CORRELATION</code>
     *        <li><code>PEARSON_ABS_CORRELATION</code>
     *        <li><code>SPEARMAN_CORRELATION</code>
     *        <li><code>KENDALL_TAU_CORRELATION</code>
     *        <li><code>EUCLIDEAN_DISTANCE</code>
     *        <li><code>CITYBLOCK_DISTANCE</code>
     *        for any other value, EUCLIDEAN_DISTANCE will be used
     *        </ul>
     * @param clusteringDimension </code>boolean</code> the clustering
     *        dimension; available dimensions:
     *        <ul>
     *        <li><code>GENES_DIMENSION</code> - the genes/rows of
     *        <code>expressionData</code> are clustered
     *        <li><code>CONDITIONS_DIMENSION</code> - the conditions/columns
     *        of <code>expressionData</code> are clustered
     *        </ul>
     * @return Node[] the set of <code>Node</code> objects, describing the
     *        hierarchical clustering solution consisting of
     *        (clusteredElements-1) nodes; depending on whether genes (rows)
     *        or conditions (columns) were clustered, clusteredElements is
     *        equal to the number of genes/rows or the number of
     *        conditions/columns, respectively
     * @throws Exception
     */
    public ClusterTreeNode[] singleLinkageClustering(double[][]
            expressionData,
            int[][] missingsMask,
            double[] elementsWeights,
            double[][] distanceMatrix,
            char distanceMetric,
            boolean clusteringDimension) throws Exception {
        int numberOfElements =
                (clusteringDimension == CONDITIONS_DIMENSION) ?
                expressionData[0].length : expressionData.length;
        int numberOfNodes = numberOfElements - 1;
        int[] vector;
        double[] temp;
        int[] indexes;
        ClusterTreeNode[] result;

        temp = new double[numberOfNodes];
        indexes = new int[numberOfElements];
        vector = new int[numberOfNodes];
        result = new ClusterTreeNode[numberOfNodes];

        for (int i = 0; i < numberOfNodes; i++) {
            vector[i] = i;
            result[i] = new ClusterTreeNode();
            result[i].setDistance(Double.MAX_VALUE);
        }

        if (distanceMatrix != null) {
            this.setSubtaskValues(expressionData.length+1);
            for (int i = 0; i < expressionData.length; i++) {
                for (int j = 0; j < i; j++) {
                    temp[j] = distanceMatrix[i][j];
                }

                for (int j = 0; j < i; j++) {
                    int k = vector[j];
                    if (result[j].getDistance() >= temp[j]) {
                        if (result[j].getDistance() < temp[k]) {
                            temp[k] = result[j].getDistance();
                        }
                        result[j].setDistance(temp[j]);
                        vector[j] = i;
                    } else if (temp[j] < temp[k]) {
                        temp[k] = temp[j];
                    }
                }

                for (int j = 0; j < i; j++) {
                    if (vector[j] < i &&
                        result[j].getDistance() >=
                        result[vector[j]].getDistance()) {
                        vector[j] = i;
                    }
                }
                this.incrementSubtaskPercentDone();
            }
        }
        else {
            this.setSubtaskValues(numberOfElements+1);
            for (int i = 0; i < numberOfElements; i++) {
                for (int j = 0; j < i; j++) {
                    temp[j] = computeDistance(expressionData,
                                              expressionData,
                                              missingsMask,
                                              missingsMask,
                                              elementsWeights,
                                              i,
                                              j,
                                              clusteringDimension,
                                              distanceMetric);
                }
                NumberFormat nf = NumberFormat.getInstance(Locale.ENGLISH);
                nf.setMaximumFractionDigits(1);
                nf.setMinimumFractionDigits(1);

                for (int j = 0; j < i; j++) {
                    int k = vector[j];
                    if (result[j].getDistance() >= temp[j]) {
                        if (result[j].getDistance() < temp[k]) {
                            temp[k] = result[j].getDistance();
                        }
                        result[j].setDistance(temp[j]);
                        vector[j] = i;
                    } else
                    if (temp[j] < temp[k]) {
                        temp[k] = temp[j];
                    }
                }

                for (int j = 0; j < i; j++) {
                    if (vector[j] < i &&
                        result[j].getDistance() >=
                        result[vector[j]].getDistance()) {
                        vector[j] = i;
                    }
                }
                this.incrementSubtaskPercentDone();
            }
        }

        for (int i = 0; i < numberOfNodes; i++) {
            result[i].setLeft(i);
        }
        Arrays.sort(result, new ClusterTreeNodeComparator());

        for (int i = 0; i < numberOfElements; i++) {
            indexes[i] = i;
        }
        for (int i = 0; i < numberOfNodes; i++) {
            int j = result[i].getLeft();
            int k = vector[j];
            result[i].setLeft(indexes[j]);
            result[i].setRight(indexes[k]);
            indexes[k] = -i - 1;
        }
        this.updatePercentDone();
        return result;
    }

    /**
     * Performs clustering using pairwise complete-linkage on the given
     * distance matrix.
     *
     * @param distanceMatrix double[][] the distance matrix; the number of
     *        rows (first dimension of this bidimensional array) is equal
     *        to the number of elements to be clustered; this is a ragged
     *        array containing the distances between the elements to be
     *        clustered (genes/rows or conditions/columns); as the distance
     *        matrix is symmetric, with zeros on the diagonal, only the
     *        lower triangular half of the distance matrix is saved and
     *        used; the distance matrix is modified by this method
     * @return Node[] the set of <code>Node</code> objects, describing the
     *        hierarchical clustering solution consisting of
     *        (clusteredElements-1) nodes; depending on whether genes (rows)
     *        or conditions (columns) were clustered, clusteredElements is
     *        equal to the number of genes/rows or the number of
     *        conditions/columns, respectively
     * @throws ArrayIndexOutOfBoundsException
     * @throws Exception
     */
    public ClusterTreeNode[] completeLinkageClustering(double[][]
            distanceMatrix) throws ArrayIndexOutOfBoundsException, Exception {
        int[] clusterIDs;
        ClusterTreeNode[] result;

        clusterIDs = new int[distanceMatrix.length];
        result = new ClusterTreeNode[distanceMatrix.length - 1];
        for (int i = 0; i < result.length; i++) {
            result[i] = new ClusterTreeNode();
        }

        /* Setup a list specifying to which cluster a gene belongs */
        for (int j = 0; j < clusterIDs.length; j++) {
            clusterIDs[j] = j;
        }

        setSubtaskValues(distanceMatrix.length-1);
        for (int n = distanceMatrix.length; n > 1; n--) {
            int[] iis = {1};
            int[] jjs = {0};
            int is;
            int js;
            result[distanceMatrix.length - n].setDistance(findClosestPair(
                    n, distanceMatrix, iis, jjs));
            is = iis[0];
            js = jjs[0];

            /* Fix the distances */
            for (int j = 0; j < js; j++) {
                distanceMatrix[js][j] =
                        Math.max(distanceMatrix[is][j], distanceMatrix[js][j]);
            }
            for (int j = js + 1; j < is; j++) {
                distanceMatrix[j][js] =
                        Math.max(distanceMatrix[is][j], distanceMatrix[j][js]);
            }
            for (int j = is + 1; j < n; j++) {
                distanceMatrix[j][js] =
                        Math.max(distanceMatrix[j][is], distanceMatrix[j][js]);
            }

            for (int j = 0; j < is; j++) {
                distanceMatrix[is][j] = distanceMatrix[n - 1][j];
            }
            for (int j = is + 1; j < n - 1; j++) {
                distanceMatrix[j][is] = distanceMatrix[n - 1][j];
            }

            /* Update clusterids */
            result[distanceMatrix.length - n].setLeft(clusterIDs[is]);
            result[distanceMatrix.length - n].setRight(clusterIDs[js]);
            clusterIDs[js] = n - distanceMatrix.length - 1;
            clusterIDs[is] = clusterIDs[n - 1];
            this.incrementSubtaskPercentDone();
        }
        this.updatePercentDone();

        return result;
    }

    /**
     * Performs clustering using pairwise average-linking on the given
     * distance matrix.
     *
     * @param distanceMatrix double[][] the distance matrix; the number of
     *        rows (first dimension of this bidimensional array) is equal
     *        to the number of elements to be clustered; this is a ragged
     *        array containing the distances between the elements to be
     *        clustered (genes/rows or conditions/columns); as the distance
     *        matrix is symmetric, with zeros on the diagonal, only the
     *        lower triangular half of the distance matrix is saved and
     *        used; the distance matrix is modified by this method
     * @return Node[] the set of <code>Node</code> objects, describing the
     *        hierarchical clustering solution consisting of
     *        (clusteredElements-1) nodes; depending on whether genes (rows)
     *        or conditions (columns) were clustered, clusteredElements is
     *        equal to the number of genes/rows or the number of
     *        conditions/columns, respectively
     * @throws ArrayIndexOutOfBoundsException
     * @throws Exception
     */
    public ClusterTreeNode[] averageLinkageClustering(double[][]
            distanceMatrix) throws ArrayIndexOutOfBoundsException, Exception {
        int numberOfClusters = distanceMatrix.length;
        int[] clusterIDs;
        int[] numberOfClustersElements;
        ClusterTreeNode[] result;

        clusterIDs = new int[numberOfClusters];
        numberOfClustersElements = new int[numberOfClusters];
        result = new ClusterTreeNode[numberOfClusters - 1];
        for (int i = 0; i < result.length; i++) {
            result[i] = new ClusterTreeNode();
        }

        /* Setup a list specifying to which cluster a gene belongs, and keep track
         * of the number of elements in each cluster (needed to calculate the
         * average). */
        for (int j = 0; j < numberOfClusters; j++) {
            numberOfClustersElements[j] = 1;
            clusterIDs[j] = j;
        }

        this.setSubtaskValues(numberOfClusters-1);
        for (int n = numberOfClusters; n > 1; n--) {
            int sum;
            int[] iis = {1};
            int[] jjs = {0};
            int is;
            int js;
            result[numberOfClusters - n].setDistance(findClosestPair(
                    n, distanceMatrix, iis, jjs));
            is = iis[0];
            js = jjs[0];

            /* Save result */
            result[numberOfClusters - n].setLeft(clusterIDs[is]);
            result[numberOfClusters - n].setRight(clusterIDs[js]);

            /* Fix the distances */
            sum = numberOfClustersElements[is] + numberOfClustersElements[js];
            for (int j = 0; j < js; j++) {
                distanceMatrix[js][j] = distanceMatrix[is][j] *
                                        numberOfClustersElements[is]
                                        + distanceMatrix[js][j] *
                                        numberOfClustersElements[js];
                distanceMatrix[js][j] /= sum;
            }
            for (int j = js + 1; j < is; j++) {
                distanceMatrix[j][js] = distanceMatrix[is][j] *
                                        numberOfClustersElements[is]
                                        + distanceMatrix[j][js] *
                                        numberOfClustersElements[js];
                distanceMatrix[j][js] /= sum;
            }
            for (int j = is + 1; j < n; j++) {
                distanceMatrix[j][js] = distanceMatrix[j][is] *
                                        numberOfClustersElements[is]
                                        + distanceMatrix[j][js] *
                                        numberOfClustersElements[js];
                distanceMatrix[j][js] /= sum;
            }

            for (int j = 0; j < is; j++) {
                distanceMatrix[is][j] = distanceMatrix[n - 1][j];
            }
            for (int j = is + 1; j < n - 1; j++) {
                distanceMatrix[j][is] = distanceMatrix[n - 1][j];
            }

            /* Update number of elements in the clusters */
            numberOfClustersElements[js] = sum;
            numberOfClustersElements[is] = numberOfClustersElements[n - 1];

            /* Update clusterids */
            clusterIDs[js] = n - numberOfClusters - 1;
            clusterIDs[is] = clusterIDs[n - 1];
            this.incrementSubtaskPercentDone();
        }
        this.updatePercentDone();
        return result;
    }

    /**
     * Performs clustering using pairwise centroid-linking on a given
     * set of gene expression data (specified by <code>expressionData</code>
     * and <code>missingsMask</code>), using the distance matrix given by
     * <code>distanceMatrix</code>.
     *
     * @param expressionData <code>double[][]</code> the data that contains
     *        the single dimension expression arrays (rows/genes or
     *        columns/conditions) to be clustered; first dimension contains
     *        genes/rows, second dimension contains conditions/columns
     * @param missingsMask <code>int[][]</code> shows which expression
     *        values are missing; if <code>missingsMask[i][j]</code> == 0,
     *        then <code>expressionData[i][j]</code> is missing; otherwise,
     *        if <code>missingsMask[i][j]</code> == 1, then
     *        <code>expressionData[i][j]</code> is valid
     * @param elementsWeights <code>double[]</code> the weights
     *        of the elements to be clustered (genes/rows or
     *        conditions/columns); the length of this array is equal to
     *        the number of conditions/columns if the distances between
     *        genes/rows are calculated, or the number of genes/rows if the
     *        distances between conditions/columns are calculated
     * @param distanceMatrix double[][] the distance matrix; if the
     *        distance matrix is not null, it is used to speed up the
     *        clustering calculation and the original expression data
     *        (specified by <code>expressionData</code> and
     *        <code>missingsMask</code>) are not needed and, therefore,
     *        ignored; the contents of the distance matrix
     *        are not modified; if <code>distanceMatrix</code> is null,
     *        the pairwise distances are calculated from the gene expression
     *        data (the <code>expressionData</code> and
     *        <code>missingsMask</code> bidimensional arrays) and stored in
     *        temporary arrays
     * @param distanceMetric <code>char</code> the distance measure to use;
     *        valid distance metrics are:<br/>
     *        <ul>
     *        <li><code>UNCENTERED_CORRELATION</code>
     *        <li><code>PEARSON_CORRELATION</code>
     *        <li><code>UNCENTERED_ABS_CORRELATION</code>
     *        <li><code>PEARSON_ABS_CORRELATION</code>
     *        <li><code>SPEARMAN_CORRELATION</code>
     *        <li><code>KENDALL_TAU_CORRELATION</code>
     *        <li><code>EUCLIDEAN_DISTANCE</code>
     *        <li><code>CITYBLOCK_DISTANCE</code>
     *        for any other value, EUCLIDEAN_DISTANCE will be used
     *        </ul>
     * @param clusteringDimension </code>boolean</code> the clustering
     *        dimension; available dimensions:
     *        <ul>
     *        <li><code>GENES_DIMENSION</code> - the genes/rows of
     *        <code>expressionData</code> are clustered
     *        <li><code>CONDITIONS_DIMENSION</code> - the conditions/columns
     *        of <code>expressionData</code> are clustered
     *        </ul>
     * @return Node[] the set of <code>Node</code> objects, describing the
     *        hierarchical clustering solution consisting of
     *        (clusteredElements-1) nodes; depending on whether genes (rows)
     *        or conditions (columns) were clustered, clusteredElements is
     *        equal to the number of genes/rows or the number of
     *        conditions/columns, respectively
     * @throws ArrayIndexOutOfBoundsException
     * @throws Exception
     */
    public ClusterTreeNode[] centroidLinkageClustering(
            double[][] expressionData,
            int[][] missingsMask,
            double[] elementsWeights,
            double[][] distanceMatrix,
            char distanceMetric,
            boolean clusteringDimension) throws ArrayIndexOutOfBoundsException,
            Exception {
        int numberOfElements = (clusteringDimension == GENES_DIMENSION) ?
                               expressionData.length : expressionData[0].length;
        int numberOfData = clusteringDimension == CONDITIONS_DIMENSION ?
                           expressionData.length : expressionData[0].length;
        int numberOfNodes = numberOfElements - 1;

        ClusterTreeNode[] result;
        double[][] newData;
        int[][] newMask;
        int[] distIDs = new int[numberOfElements];
        result = new ClusterTreeNode[numberOfNodes];
        for (int i = 0; i < result.length; i++) {
            result[i] = new ClusterTreeNode();
        }

        newData = new double[numberOfElements][numberOfData];
        newMask = new int[numberOfElements][numberOfData];
        for (int i = 0; i < numberOfElements; i++) {
            distIDs[i] = i;
        }
        /* To remember which row/column in the distance matrix contains what */

        /* Storage for node data */
        if (clusteringDimension == CONDITIONS_DIMENSION) {
            for (int i = 0; i < numberOfElements; i++) {
                for (int j = 0; j < numberOfData; j++) {
                    newData[i][j] = expressionData[j][i];
                    newMask[i][j] = missingsMask[j][i];
                }
            }
            expressionData = newData;
            missingsMask = newMask;
        } else {
            for (int i = 0; i < numberOfElements; i++) {
                for (int j = 0; j < numberOfData; j++) {
                    newData[i][j] = expressionData[i][j];
                    newMask[i][j] = missingsMask[i][j];
                }
            }
            expressionData = newData;
            missingsMask = newMask;
        }

        this.setSubtaskValues(numberOfNodes);
        for (int inode = 0; inode < numberOfNodes; inode++) {
            /* Find the pair with the shortest distance */
            int[] iis = {1};
            int[] jjs = {0};
            int is;
            int js;
            result[inode].setDistance(findClosestPair(
                    numberOfElements - inode, distanceMatrix, iis, jjs));
            is = iis[0];
            js = jjs[0];
            result[inode].setLeft(distIDs[js]);
            result[inode].setRight(distIDs[is]);

            /* Make node js the new node */
            for (int i = 0; i < numberOfData; i++) {
                expressionData[js][i] = expressionData[js][i] *
                                        missingsMask[js][i] +
                                        expressionData[is][i] *
                                        missingsMask[is][i];
                missingsMask[js][i] += missingsMask[is][i];
                if (missingsMask[js][i] != 0) {
                    expressionData[js][i] = expressionData[js][i] * 1.0 /
                                            missingsMask[js][i];
                }
            }
            expressionData[is] = expressionData[numberOfNodes - inode];
            missingsMask[is] = missingsMask[numberOfNodes - inode];

            /* Fix the distances */
            distIDs[is] = distIDs[numberOfNodes - inode];
            for (int i = 0; i < is; i++) {
                distanceMatrix[is][i] = distanceMatrix[numberOfNodes - inode][i];
            }
            for (int i = is + 1; i < numberOfNodes - inode; i++) {
                distanceMatrix[i][is] = distanceMatrix[numberOfNodes - inode][i];
            }

            distIDs[js] = -inode - 1;
            for (int i = 0; i < js; i++) {
                distanceMatrix[js][i] = computeDistance(expressionData,
                        expressionData,
                        missingsMask,
                        missingsMask,
                        elementsWeights,
                        js,
                        i,
                        false,
                        distanceMetric);
            }
            for (int i = js + 1; i < numberOfNodes - inode; i++) {
                distanceMatrix[i][js] = computeDistance(expressionData,
                        expressionData,
                        missingsMask,
                        missingsMask,
                        elementsWeights,
                        js,
                        i,
                        false,
                        distanceMetric);
            }
            this.incrementSubtaskPercentDone();
        }
        this.updatePercentDone();

        return result;
    }

    /**
     * Calculates the distance between two genes/rows or two
     * conditions/columns. To compute the distance between two
     * genes/rows or conditions/columns of the same expression
     * data, just call this method with the same expression matrix
     * for <code>expressionData1</code> and <code>expressionData2</code>
     * and the corresponding mask in both <code>missingsMask1</code>
     * and <code>missingsMask2</code>.<br/>
     *
     * <code>expressionData1</code>, <code>expressionData2</code>,
     * <code>missingsMask1</code> and <code>missingsMask2</code> must
     * have the same dimensions.
     *
     * @param expressionData1 <code>double[][]</code> the expression
     *        data matrix containing the first gene/row or condition/column
     * @param expressionData2 <code>double[][]</code> the expression
     *        data matrix containing the second gene/row or condition/column
     * @param missingsMask1 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData1</code>
     * @param missingsMask2 <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData2</code>
     * @param elementsWeights <code>double[]</code> the weights
     *        of the elements to be clustered (genes/rows or
     *        conditions/columns); the length of this array is equal to
     *        the number of conditions/columns if the distances between
     *        genes/rows are calculated, or the number of genes/rows if the
     *        distances between conditions/columns are calculated
     * @param index1 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData1</code> and
     *        <code>missingsMask1</code>
     * @param index2 <code>int</code> the index of the first gene/row or
     *        condition/column in <code>expressionData2</code> and
     *        <code>missingsMask2</code>
     * @param dimension </code>boolean</code> the dimension in which
     *        distance is being measured; available dimensions:
     *        <ul>
     *        <li><code>GENES_DIMENSION</code> - the distance is measured
     *        between two genes/rows
     *        <li><code>CONDITIONS_DIMENSION</code> - the distance is
     *        measured  between two conditions/columns
     *        </ul>
     * @param distanceMetric <code>char</code> the distance measure to use;
     *        valid distance metrics are:<br/>
     *        <ul>
     *        <li><code>UNCENTERED_CORRELATION</code>
     *        <li><code>PEARSON_CORRELATION</code>
     *        <li><code>UNCENTERED_ABS_CORRELATION</code>
     *        <li><code>PEARSON_ABS_CORRELATION</code>
     *        <li><code>SPEARMAN_CORRELATION</code>
     *        <li><code>KENDALL_TAU_CORRELATION</code>
     *        <li><code>EUCLIDEAN_DISTANCE</code>
     *        <li><code>CITYBLOCK_DISTANCE</code>
     *        for any other value, EUCLIDEAN_DISTANCE will be used
     *        </ul>
     * @return double the computed distance
     * @throws Exception
     */
    public static double computeDistance(double[][] expressionData1,
                                         double[][] expressionData2,
                                         int[][] missingsMask1,
                                         int[][] missingsMask2,
                                         double[] elementsWeights,
                                         int index1,
                                         int index2,
                                         boolean dimension,
                                         char distanceMetric) throws Exception {
        switch (distanceMetric) {
        case EUCLIDEAN_DISTANCE:
            return euclideanDistance(expressionData1, expressionData2,
                                     missingsMask1, missingsMask2,
                                     elementsWeights,
                                     index1, index2, dimension);
        case CITYBLOCK_DISTANCE:
            return cityblockDistance(expressionData1, expressionData2,
                                     missingsMask1, missingsMask2,
                                     elementsWeights,
                                     index1, index2, dimension);
        case PEARSON_CORRELATION:
            return pearsonsCorrelation(expressionData1, expressionData2,
                                       missingsMask1, missingsMask2,
                                       elementsWeights,
                                       index1, index2, dimension);
        case PEARSON_ABS_CORRELATION:
            return pearsonsAbsoluteCorrelation(expressionData1, expressionData2,
                                               missingsMask1, missingsMask2,
                                               elementsWeights,
                                               index1, index2, dimension);
        case UNCENTERED_CORRELATION:
            return uncenteredCorrelation(expressionData1, expressionData2,
                                         missingsMask1, missingsMask2,
                                         elementsWeights,
                                         index1, index2, dimension);
        case UNCENTERED_ABS_CORRELATION:
            return uncenteredAbsoluteCorrelation(expressionData1,
                                                 expressionData2,
                                                 missingsMask1, missingsMask2,
                                                 elementsWeights,
                                                 index1, index2, dimension);
        case SPEARMAN_RANK_CORRELATION:
            return spearmansRankCorrelation(expressionData1, expressionData2,
                                            missingsMask1, missingsMask2,
                                            elementsWeights,
                                            index1, index2, dimension);
        case KENDALL_TAU_CORRELATION:
            return kendallsTauCorrelation(expressionData1, expressionData2,
                                          missingsMask1, missingsMask2,
                                          elementsWeights,
                                          index1, index2, dimension);
        default:
            return euclideanDistance(expressionData1, expressionData2,
                                     missingsMask1, missingsMask2,
                                     elementsWeights,
                                     index1, index2, dimension);
        }
    }

    /**
     * Searches the distance matrix to find the pair with the shortest
     * distance between them. The row and column indexes of the pair
     * are returned in <code>ipointer</code> and <code>jpointer</code>
     * (should be acessible at ipointer[0] and jpointer[0] to be precise),
     * respectively, and the distance itself is the return value.
     *
     * @param maxRowLimit <code>int</code> the maximum number of elements
     *        of the distance matrix where this method should look for the
     *        pair with the shortest distance; this value should not exceed
     *        the number of rows in the <code>distanceMatrix</code>
     * @param distanceMatrix double[][] the distance matrix; this is a ragged
     *        array containing the distances between elements (genes/rows or
     *        conditions/columns); as the distance matrix is symmetric,
     *        with zeros on the diagonal, only the lower triangular half
     *        of the distance matrix is saved and used
     * @param ipointer <code>int[]</code> an array with, at least, one
     *        element to receive the first index of the pair with the
     *        shortest distance
     * @param jpointer <code>int[]</code> an array with, at least, onde element
     *        to receive the second index of the pair with the shortest distance
     * @return double the shortest distance between two elements
     * @throws Exception
     */
    public static double findClosestPair(int maxRowLimit,
                                         double[][] distanceMatrix,
                                         int[] ipointer,
                                         int[] jpointer) throws Exception {
        if (maxRowLimit > distanceMatrix.length) {
            maxRowLimit = distanceMatrix.length;
        }
        if (distanceMatrix == null) {
            throw new Exception(
                    "Distance matrix is empty. Can't find closest pair.");
        }
        if (ipointer == null || jpointer == null) {
            throw new Exception("Two arrays/pointers are needed with, " +
                                "at least, one element to store the indexes " +
                                "of the pair with the shortest distance.");
        }
        try {
            double temp;
            double distance = distanceMatrix[1][0];
            ipointer[0] = 1;
            jpointer[0] = 0;
            for (int i = 1; i < maxRowLimit; i++) {
                for (int j = 0; j < i; j++) {
                    temp = distanceMatrix[i][j];
                    if (temp < distance) {
                        distance = temp;
                        ipointer[0] = i;
                        jpointer[0] = j;
                    }
                }
            }
            return distance;
        } catch (ArrayIndexOutOfBoundsException ex) {
            ex.printStackTrace();
            throw new ArrayIndexOutOfBoundsException("This array is not a " +
                    "valid ragged array which contains a distance matrix.");
        }
    }

    /**
     * This function calculates the weights using the weighting scheme
     * proposed by Michael Eisen:
     * w[i] = 1.0 / sum_{j where d[i][j]<cutoff} (1 - d[i][j]/cutoff)^exponent
     * where the cutoff and the exponent are specified by the user.
     *
     * @param expressionData <code>double[][]</code> the bidimensional
     *        array that contains the expression data
     * @param missingsMask <code>int[][]</code> the mask that indicates
     *        which expression values are missing in <code>expressionData</code>
     * @param originalWeights <code>double[]</code> the original weights
     *        used to calculate the distances
     * @param dimension <code>boolean</code> the dimension of the computation
     *        (either <code>GENES_DIMENSION</code> or
     *        <code>CONDITIONS_DIMENSION</code>)
     * @param distanceMetric <code>char</code> the distance metric used
     *        to calculate distances
     * @param cutOff <code>double</code> the cutoff to use in weight calculation
     * @param exponent <code>double</code> the exponent to use in weight calculation
     * @return double[] the computed genes or conditions weights
     * @throws Exception
     */
    public static double[] calculateWeights(double[][] expressionData,
                                            int[][] missingsMask,
                                            double[] originalWeights,
                                            boolean dimension,
                                            char distanceMetric,
                                            double cutOff,
                                            double exponent) throws Exception {
        int numberOfElements = (dimension == GENES_DIMENSION) ?
                               expressionData.length : expressionData[0].length;

        double[] result = new double[numberOfElements];
        for (int i = 0; i < numberOfElements; i++) {
            result[i] = 0;
        }

        for (int i = 0; i < numberOfElements; i++) {
            result[i] += 1.0;
            for (int j = 0; j < i; j++) {
                double distance = computeDistance(expressionData,
                                                  expressionData,
                                                  missingsMask,
                                                  missingsMask,
                                                  originalWeights,
                                                  i, j, dimension,
                                                  distanceMetric);
                if (distance < cutOff) {
                    double dweight = Math.exp(exponent *
                                              Math.log(1 - distance / cutOff));

                    result[i] += dweight;
                    result[j] += dweight;
                }
            }
        }
        for (int i = 0; i < numberOfElements; i++) {
            result[i] = 1.0 / result[i];
        }
        return result;
    }
}
