package biggests.gui;

import java.awt.*;
import java.awt.event.*;
import java.io.*;

import javax.swing.*;

import biggests.clustering.*;
import edu.stanford.genetics.treeview.*;
import edu.stanford.genetics.treeview.app.*;
import smadeira.biclustering.*;
import java.util.Vector;

/**
 * <p>Title: Panel for Analyzing Matrix Dendrograms</p>
 *
 * <p>Description: Allows the user to parameterize and apply
 *                 hierarchical clustering algorithms to obtain a
 *                 hierarchical representation of the gene expression
 *                 data, a dendrogram.</p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana Gonalves
 */
public class PanelAnalyzingMatrixDendrogram extends JPanel {
    public static final String hierarchicalClusteringTempFilesDirectoryPath =
            "hc-temp";

    /**
     * Similarity metrics (names to display).
     */
    public static final String[] distanceMetrics = {"Uncentered correlation",
            "Pearson's correlation",
            "Uncentered absolute correlation",
            "Pearson's absolute correlation",
            "Spearman's rank correlation",
            "Kendall's tau correlation",
            "Euclidean distance",
            "Cityblock distance"};

    /**
     * Similarity metrics alias (the arguments to pass to
     * hierarchical methods). Each alias corresponds to the
     * <code>distanceMetrics</code> element with the same
     * index.
     */
    public static final char[] distanceMetricsAlias = {
            HierarchicalClustering.UNCENTERED_CORRELATION,
            HierarchicalClustering.PEARSON_CORRELATION,
            HierarchicalClustering.UNCENTERED_ABS_CORRELATION,
            HierarchicalClustering.PEARSON_ABS_CORRELATION,
            HierarchicalClustering.SPEARMAN_RANK_CORRELATION,
            HierarchicalClustering.KENDALL_TAU_CORRELATION,
            HierarchicalClustering.EUCLIDEAN_DISTANCE,
            HierarchicalClustering.CITYBLOCK_DISTANCE};

    /**
     * Clustering methods (names to display).
     */
    public static final String[] clusteringMethods = {"Centroid linkage",
            "Single linkage",
            "Complete linkage",
            "Average linkage"};

    /**
     * Clustering methods alias (the arguments to pass to
     * hierarchical methods). Each alias corresponds to the
     * <code>clusteringMethods</code> element with the same
     * index.
     */
    public static final char[] clusteringMethodsAlias = {
            HierarchicalClustering.CENTROID_LINKAGE,
            HierarchicalClustering.SINGLE_LINKAGE,
            HierarchicalClustering.COMPLETE_LINKAGE,
            HierarchicalClustering.AVERAGE_LINKAGE};

    /**
     * Hierarchical clustering label.
     */
    private JLabel hierarchicalClusteringLabel;

    /**
     * Cluster genes check box. Allows the user to choose if he/she wants
     * to perform clustering on the genes dimension.
     */
    private JCheckBox clusterGenesCheckBox;

    /**
     * Cluster conditions check box. Allows the user to choose if he/she wants
     * to perform clustering on the conditions dimension.
     */
    private JCheckBox clusterConditionsCheckBox;

    /**
     * Genes distance metric label.
     */
    private JLabel genesDistanceMetricLabel;

    /**
     * Conditions distance metric label.
     */
    private JLabel conditionsDistanceMetricLabel;

    /**
     * Genes distance metric combo box. Allows the user to select
     * which similarity metric to use in genes dimension clustering.
     */
    private JComboBox genesDistanceMetricComboBox;

    /**
     * Conditions distance metric combo box. Allows the user to select
     * which similarity metric to use in conditions dimension clustering.
     */
    private JComboBox conditionsDistanceMetricComboBox;

    /**
     * Clustering methods list. Allows the user to select which
     * clustering method to use.
     */
    private JList clusteringMethodList;

    /**
     * Clustering methods label.
     */
    private JLabel clusteringMethodLabel;

    /**
     * The layout manager for this panel.
     */
    private SpringLayout springLayout;

    /**
     * The apply button. When the user clicks it, hierarchical
     * clustering is performed and the resulting dendrogram is
     * presented in a new window.
     */
    private JButton applyButton;

    /**
     * The hierarchical clustering object. It is initialized by
     * <code>paint</code> with a given expression data matrix,
     * over which hierarchical clustering will be performed.
     */
    private HierarchicalClustering hierarchicalClustering;

    /**
     * The application's main frame.
     */
    private BiggestsFrame frame;

    /**
     * A Java TreeView application.
     */
    private MyTreeViewApp treeViewApp;

    /**
     * The number of dendrograms shown until this moment.
     */
    private static int numberOfDendrograms;

    /**
     * Creates a new PanelAnalyzingMatrixDendrogram with options
     * to perform hierarchical clustering over a set of expression
     * data.
     *
     * @param frame the <code>BiggestsFrame</code>
     */
    public PanelAnalyzingMatrixDendrogram(BiggestsFrame frame) {
        this.frame = frame;
        try {
            jbInit();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    /**
     * Initializes graphical components.
     *
     * @throws Exception if components cannot be initialized
     */
    private void jbInit() throws Exception {
        this.setPreferredSize(new Dimension(550, 350));
        this.setBorder(BorderFactory.createEmptyBorder(20, 20, 20, 20));
        springLayout = new SpringLayout();
        hierarchicalClusteringLabel = new JLabel("Hierarchical Clustering");
        hierarchicalClusteringLabel.setFont(new Font(
                hierarchicalClusteringLabel.getFont().getName(),
                Font.BOLD,
                16));
        clusterGenesCheckBox = new JCheckBox("Cluster genes");
        clusterConditionsCheckBox = new JCheckBox("Cluster time points");
        genesDistanceMetricLabel = new JLabel("Similarity metric");
        conditionsDistanceMetricLabel = new JLabel("Similarity metric");
        clusteringMethodList = new JList(clusteringMethods);
        clusteringMethodList.setSelectedIndex(2);
        clusteringMethodList.setSelectionMode(ListSelectionModel.
                                              SINGLE_SELECTION);
        clusteringMethodLabel = new JLabel("Clustering method");
        genesDistanceMetricComboBox = new JComboBox(distanceMetrics);
        conditionsDistanceMetricComboBox = new JComboBox(distanceMetrics);
        applyButton = new JButton("View Dendrogram");
        applyButton.setPreferredSize(new Dimension(470, 50));
        applyButton.addActionListener(new
                PanelAnalyzingMatrixDendrogram_applyButton_actionAdapter(this));
        this.setLayout(springLayout);
        clusterGenesCheckBox.addItemListener(
            new PanelAnalyzingMatrixDendrogram_clusterGenesCheckBox_itemAdapter(this));
        clusterConditionsCheckBox.addItemListener(
            new PanelAnalyzingMatrixDendrogram_clusterConditionsCheckBox_itemAdapter(this));

        this.clusterGenesCheckBox.setSelected(true);
        this.conditionsDistanceMetricLabel.setEnabled(false);
        this.conditionsDistanceMetricComboBox.setEnabled(false);

        this.add(hierarchicalClusteringLabel);

        this.add(clusterGenesCheckBox);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   clusterGenesCheckBox, 15,
                                   SpringLayout.SOUTH,
                                   hierarchicalClusteringLabel);
        springLayout.putConstraint(SpringLayout.WEST,
                                   clusterGenesCheckBox, 0,
                                   SpringLayout.WEST,
                                   hierarchicalClusteringLabel);

        this.add(genesDistanceMetricLabel);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   genesDistanceMetricLabel, 5,
                                   SpringLayout.SOUTH,
                                   clusterGenesCheckBox);
        springLayout.putConstraint(SpringLayout.WEST,
                                   genesDistanceMetricLabel, 0,
                                   SpringLayout.WEST,
                                   clusterGenesCheckBox);

        this.add(genesDistanceMetricComboBox);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   genesDistanceMetricComboBox, 5,
                                   SpringLayout.SOUTH,
                                   genesDistanceMetricLabel);
        springLayout.putConstraint(SpringLayout.WEST,
                                   genesDistanceMetricComboBox, 0,
                                   SpringLayout.WEST,
                                   genesDistanceMetricLabel);

        this.add(clusterConditionsCheckBox);
        springLayout.putConstraint(SpringLayout.SOUTH,
                                   clusterConditionsCheckBox, 0,
                                   SpringLayout.SOUTH,
                                   clusterGenesCheckBox);
        springLayout.putConstraint(SpringLayout.WEST,
                                   clusterConditionsCheckBox, 150,
                                   SpringLayout.EAST,
                                   clusterGenesCheckBox);

        this.add(conditionsDistanceMetricLabel);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   conditionsDistanceMetricLabel, 5,
                                   SpringLayout.SOUTH,
                                   clusterConditionsCheckBox);
        springLayout.putConstraint(SpringLayout.WEST,
                                   conditionsDistanceMetricLabel, 0,
                                   SpringLayout.WEST,
                                   clusterConditionsCheckBox);

        this.add(conditionsDistanceMetricComboBox);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   conditionsDistanceMetricComboBox, 5,
                                   SpringLayout.SOUTH,
                                   conditionsDistanceMetricLabel);
        springLayout.putConstraint(SpringLayout.WEST,
                                   conditionsDistanceMetricComboBox, 0,
                                   SpringLayout.WEST,
                                   conditionsDistanceMetricLabel);

        this.add(clusteringMethodLabel);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   clusteringMethodLabel, 15,
                                   SpringLayout.SOUTH,
                                   genesDistanceMetricComboBox);
        springLayout.putConstraint(SpringLayout.WEST,
                                   clusteringMethodLabel, 180,
                                   SpringLayout.WEST,
                                   genesDistanceMetricComboBox);

        this.add(clusteringMethodList);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   clusteringMethodList, 5,
                                   SpringLayout.SOUTH,
                                   clusteringMethodLabel);
        springLayout.putConstraint(SpringLayout.WEST,
                                   clusteringMethodList, 0,
                                   SpringLayout.WEST,
                                   clusteringMethodLabel);

        this.add(applyButton);
        springLayout.putConstraint(SpringLayout.NORTH,
                                   applyButton, 15,
                                   SpringLayout.SOUTH,
                                   clusteringMethodList);
        springLayout.putConstraint(SpringLayout.WEST,
                                   applyButton, 0,
                                   SpringLayout.WEST,
                                   genesDistanceMetricComboBox);
    }

    private String generateNewJobName() {
        numberOfDendrograms++;
        return ("hierarchical_clustering" + numberOfDendrograms);
    }

    public static void setNumberOfDendrograms(int numDend) {
        numberOfDendrograms = numDend;
    }

    public static int getNumberOfDendrograms() {
        return numberOfDendrograms;
    }

    public static void deleteTempFilesAndDirectory() {
        File output = new File(hierarchicalClusteringTempFilesDirectoryPath);
        if (output.exists()) {
            if (output.isFile()) {
                output.delete();
                return;
            }
            if (output.isDirectory()) {
                BiggestsIO.deleteFolderContent(output.getAbsolutePath());
                output.delete();
            }
        }
    }

    public void endTreeViewApp() {
        if (treeViewApp != null) {
            treeViewApp.endProgram();
            treeViewApp = null;
        }
        deleteTempFilesAndDirectory();
    }

    /**
     * Clears this panel. Sets default values for selected options.
     */
    public void clear() {
        /*        if (treeViewApp != null) {
                    int option = JOptionPane.showOptionDialog(this,
                            "Close Java TreeView?",
                            "Close Java TreeView?",
                            JOptionPane.YES_NO_OPTION,
                            JOptionPane.QUESTION_MESSAGE, null, null,
                            JOptionPane.YES_OPTION);
                    if (option == JOptionPane.YES_OPTION) {
                        this.endTreeViewApp();
                    }
                }
         */
        hierarchicalClustering = null;
        this.clusterGenesCheckBox.setSelected(true);
        this.genesDistanceMetricLabel.setEnabled(true);
        this.genesDistanceMetricComboBox.setEnabled(true);
        this.clusterConditionsCheckBox.setSelected(false);
        this.conditionsDistanceMetricLabel.setEnabled(false);
        this.conditionsDistanceMetricComboBox.setEnabled(false);
        this.clusteringMethodLabel.setEnabled(true);
        this.clusteringMethodList.setEnabled(true);
        this.applyButton.setEnabled(true);
    }

    /**
     * Paints this panel to perform hierarchical clustering over
     * an expression matrix.
     *
     * @param matrix the <code>IMatrix</code> to perform hierarchical
     *        clustering on
     */
    public void paint(IMatrix matrix) {
        hierarchicalClustering = new HierarchicalClustering(matrix);
    }

    /**
     * Handles the event fired when the item state of cluster
     * genes check box changes.
     *
     * @param e the <code>ItemEvent</code> fired when the item state
     *        of cluster genes check box changes
     */
    public void clusterGenesCheckBox_itemStateChanged(ItemEvent e) {
        this.genesDistanceMetricLabel.setEnabled(clusterGenesCheckBox.
                                                 isSelected());
        this.genesDistanceMetricComboBox.setEnabled(clusterGenesCheckBox.
                isSelected());
        boolean enabled = true;
        if (!clusterGenesCheckBox.isSelected() &&
            !clusterConditionsCheckBox.isSelected()) {
            enabled = false;
        }
        this.clusteringMethodLabel.setEnabled(enabled);
        this.clusteringMethodList.setEnabled(enabled);
        this.applyButton.setEnabled(enabled);
    }

    /**
     * Handles the event fired when the item state of cluster
     * conditions check box changes.
     *
     * @param e the <code>ItemEvent</code> fired when the item state
     *        of cluster conditions check box changes
     */
    public void clusterConditionsCheckBox_itemStateChanged(ItemEvent e) {
        this.conditionsDistanceMetricLabel.setEnabled(clusterConditionsCheckBox.
                isSelected());
        this.conditionsDistanceMetricComboBox.setEnabled(
                clusterConditionsCheckBox.isSelected());
        boolean enabled = true;
        if (!clusterGenesCheckBox.isSelected() &&
            !clusterConditionsCheckBox.isSelected()) {
            enabled = false;
        }
        this.clusteringMethodLabel.setEnabled(enabled);
        this.clusteringMethodList.setEnabled(enabled);
        this.applyButton.setEnabled(enabled);
    }

    /**
     * Handles the event fired when the user clicks apply button.
     * Performs hierarchical clustering over the expression data with
     * the selected options and opens a new Java Treeview window with
     * the resulting dendrogram. If no clustering (genes or conditions)
     * is selected, hierarchical clustering won't be performed.
     *
     * @param e <code>ActionEvent</code> fired when the user clicks
     *        apply button
     */
    public void applyButton_actionPerformed(ActionEvent e) {
        if (!applyButton.isEnabled() || this.hierarchicalClustering == null) {
            return;
        }
        char genesMetric = '\0';
        char conditionsMetric = '\0';
        if (clusterGenesCheckBox.isSelected()) {
            genesMetric = distanceMetricsAlias[genesDistanceMetricComboBox.
                          getSelectedIndex()];
        }
        if (clusterConditionsCheckBox.isSelected()) {
            conditionsMetric = distanceMetricsAlias[
                               conditionsDistanceMetricComboBox.
                               getSelectedIndex()];
        }

        if (genesMetric == '\0' && conditionsMetric == '\0') {
            return;
        }

        char clusteringMethod = clusteringMethodsAlias[clusteringMethodList.
                                getSelectedIndex()];

        String jobname = this.generateNewJobName();
        File output = new File(hierarchicalClusteringTempFilesDirectoryPath);
        if (output.exists()) {
            if (output.isFile()) {
                output.delete();
                output.mkdirs();
            } else
            if (output.isDirectory()) {
                BiggestsIO.deleteFolderContent(output.getAbsolutePath());
            }
        } else {
            output.mkdirs();
        }

        ThreadHierarchicalClustering thread =
                new ThreadHierarchicalClustering(frame, hierarchicalClustering,
                                                 genesMetric, conditionsMetric,
                                                 clusteringMethod,
                                                 output.getName(), jobname);
    }

    public void showDendrogram(String cdtDirectory, String cdtFileName) {
        if (treeViewApp == null) {
            treeViewApp = new MyTreeViewApp();
        }
        File output = new File(cdtDirectory + File.separator + cdtFileName);
        try {
            System.out.println(cdtDirectory + File.separator + cdtFileName);
            ViewFrame viewFrame = treeViewApp.openNew(new FileSet(output.
                    getName(),
                    output.getParent() + File.separator));
            viewFrame.setVisible(true);
        } catch (Exception ex) {
            JOptionPane.showMessageDialog(this, "Unable to load dendrogram.",
                                          "Unable to load dendrogram",
                                          JOptionPane.ERROR_MESSAGE);
            return;
        }
    }
}

class PanelAnalyzingMatrixDendrogram_applyButton_actionAdapter implements
        ActionListener {
    private PanelAnalyzingMatrixDendrogram adaptee;
    PanelAnalyzingMatrixDendrogram_applyButton_actionAdapter(
            PanelAnalyzingMatrixDendrogram adaptee) {
        this.adaptee = adaptee;
    }

    public void actionPerformed(ActionEvent e) {
        adaptee.applyButton_actionPerformed(e);
    }
}


class PanelAnalyzingMatrixDendrogram_clusterConditionsCheckBox_itemAdapter implements
        ItemListener {
    private PanelAnalyzingMatrixDendrogram adaptee;
    PanelAnalyzingMatrixDendrogram_clusterConditionsCheckBox_itemAdapter(
            PanelAnalyzingMatrixDendrogram adaptee) {
        this.adaptee = adaptee;
    }

    public void itemStateChanged(ItemEvent e) {
        adaptee.clusterConditionsCheckBox_itemStateChanged(e);
    }
}


class PanelAnalyzingMatrixDendrogram_clusterGenesCheckBox_itemAdapter implements
        ItemListener {
    private PanelAnalyzingMatrixDendrogram adaptee;
    PanelAnalyzingMatrixDendrogram_clusterGenesCheckBox_itemAdapter(
            PanelAnalyzingMatrixDendrogram adaptee) {
        this.adaptee = adaptee;
    }

    public void itemStateChanged(ItemEvent e) {
        adaptee.clusterGenesCheckBox_itemStateChanged(e);
    }
}


/**
 *
 * <p>Title: </p>
 *
 * <p>Description: </p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 2 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/copyleft/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonalves
 */
class MyTreeViewApp extends LinkedViewApp {
    /**
     * Creates a new MyTreeViewApp to present results of
     * hierarchical clustering over a set of expression data.
     */
    public MyTreeViewApp() {
        super();
    }

    /**
     * Constructor for MyTreeViewApp object.
     * Takes configuration from <code>xmlConfig</code>.
     *
     * @param xmlConfig <code>XmlConfig</code> the configuration file
     */
    public MyTreeViewApp(XmlConfig xmlConfig) {
        super(xmlConfig);
    }

    /**
     * Overrides <code>LinkedViewApp</code>'s <code>windowClosed</code>
     * method.
     *
     * @param e <code>WindowEvent</code>
     */
    public void windowClosed(WindowEvent e) {
        windows.removeElement(e.getWindow());
        if (windows.isEmpty()) {
            endProgram();
            return;
        }

        rebuildWindowMenus();
    }

    /**
     * Reimplements <code>LinkedViewApp</code>'s <code>endProgram</code>
     * method (the original method would call System.exit(0) and cause
     * BiGGEsTS application to terminate, which we don't want to happen here).
     */
    protected void endProgram() {
        try {
            if (this.getGlobalConfig() != null) {
                this.getGlobalConfig().store();
            }
            closeAllWindows();
        } catch (java.security.AccessControlException e) {
            //we're probably running in a applet here...
        }
    }
}
