package biggests.gui;

import java.io.*;

import javax.swing.*;
import javax.swing.tree.*;

import biggests.utils.*;
import smadeira.biclustering.*;
import smadeira.ontologizer.*;

/**
 * <p>Title: Biclustering and function analysis thread.</p>
 *
 * <p>Description: Thread to apply biclustering algorithms and function analysis
 *                (long time consuming task) while displaying a progress
 *                bar dialog to the user.</p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonalves
 */
public class ThreadBiclustering extends Thread {
    /**
     * Biclusters group object.
     */
    private Biclustering biclustering;

    /**
     * The dataset tree selected path which contains the node
     * information object of the matrix that is being biclustered.
     */
    private TreePath selectionPath;

    /**
     * Progress bar dialog.
     */
    private ProgressBarFrame progressFrame;

    /**
     * Application's main frame.
     */
    private BiggestsFrame mainFrame;

    /**
     * Creates a new <code>ThreadBiclusteringFunctionAnalysis</code>
     * @param frame the BiGGEsTS main window
     * @param biclustering <code>Biclustering</code> object to apply
     *        biclustering algorithm and function analysis on to
     * @param selectionPath the selected <code>TreePath</code> (path
     *        to biclustering node in the dataset tree)
     */
    public ThreadBiclustering(
            BiggestsFrame frame,
            Biclustering biclustering,
            TreePath selectionPath) {
        super("Biclustering Function Analysis Thread");

        this.biclustering = biclustering;
        this.selectionPath = selectionPath;
        mainFrame = frame;

        progressFrame =
                new ProgressBarFrame(mainFrame, this,
                                     "preparing to apply algorithm...", true);
        this.start();
    }

    /**
     * Overrides <code>Thread</code>'s  <code>run</code> method. Performs the
     * actual task of this working thread (applying the required biclustering
     * algorithm and performing function analysis on a given biclusters group).
     *
     * @see Thread#run()
     */
    public void run() {
        try {
            GOFrontEnd go = this.computeBiclusters();
            progressFrame.setString("adding new biclusters group to tree...");
            mainFrame.getPanelBiclustering().setStringsAndStyles(biclustering);
            mainFrame.getPanelBiclustering().addBiclusterGroupToDatasetTree(
                    biclustering, selectionPath, go);
            biclustering.setTaskPercentages(null);
            biclustering.setProgressBar(null);

            progressFrame.dispose();
            mainFrame.setEnabled(true);
        } catch (Exception e) {
            e.printStackTrace();
            this.biclusteringError(e.getMessage());
        } catch (StackOverflowError e) {
            e.printStackTrace();
            this.biclusteringError(e.getMessage());
        } catch (OutOfMemoryError e) {
            e.printStackTrace();
            this.biclusteringError("Unable to extract biclusters.\nOut of memory.");
        }
    }

    /**
     * Computes biclusters and GO frontend for function analysis.
     *
     * @return GOFrontEnd the GO frontend for function analysis
     * @throws Exception if biclusters cannot be computed
     * @throws StackOverflowError if there is a stack overflow
     * @throws OutOfMemoryError if runs out of memory
     */
    public GOFrontEnd computeBiclusters() throws Exception, StackOverflowError,
            OutOfMemoryError {
        biclustering.setProgressBar(progressFrame.getProgressBar());
        if (biclustering instanceof CCC_Biclustering) {
            progressFrame.setString("computing CCC-Biclusters...");
            ((CCC_Biclustering) biclustering).computeBiclusters(true);
        }
        else
        if (biclustering instanceof E_CCC_Biclustering) {
            progressFrame.setString("computing e-CCC-Biclusters...");
            ((E_CCC_Biclustering) biclustering).computeBiclusters(true);
        }
        else {
            progressFrame.setString("computing CC-TSB-biclusters...");
            ((CC_TSB_Biclustering) biclustering).computeBiclusters();
        }
        progressFrame.setString("computing GO frontend for function analysis...");
        GOFrontEnd go = mainFrame.getPanelAnalyzingFunctionTable().
                        calculateGOFrontEnd(biclustering);
        return go;
    }

    /**
     * Displays an error message when something went wrong with
     * biclustering algorithm.
     *
     * @param message String
     */
    public void biclusteringError(String message) {
        biclustering = null;
        progressFrame.dispose();
        mainFrame.setEnabled(true);
        JOptionPane.showMessageDialog(mainFrame,
                                      "Unable to apply biclustering algorithm." +
                                      "\n" + message + ".",
                                      "Unable to compute biclusters",
                                      JOptionPane.ERROR_MESSAGE);
    }
}
