package biggests.gui;

import java.io.*;
import java.util.*;

import javax.swing.*;
import javax.swing.tree.*;

import biggests.utils.*;
import smadeira.biclustering.*;
import smadeira.utils.TaskPercentages;

/**
 * <p>Title: Post-Processing Thread</p>
 *
 * <p>Description: Implements a thread to apply post-processing algorithms,
 *                 such as filtering and sorting of biclusters based on
 *                 well known metrics. Also informs a ProgressBarFrame
 *                 about the progress of the post-processing operations.
 * </p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonalves
 */
public class ThreadPostProcessing extends Thread {
    /**
     * The post-processing panel which makes the user interface
     * for post-processing options.
     */
    PanelPostProcessing postProcessingPanel;

    /**
     * The selected path in the dataset tree, holding the bicluster
     * group object to post-process.
     */
    TreePath selectionPath;

    /**
     * The bicluster group to post-process.
     */
    Biclustering biclustering;

    /**
     * The frame which holds the bar that displays information
     * about the post-processing algorithms application progress.
     */
    ProgressBarFrame progressFrame;

    /**
     * Creates a new post-processing thread, to post-process a
     * group of biclusters.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to post-process
     * @param selectionPath the selected <code>TreePath</code> in
     *        the dataset tree, which contains the <code>biclustering</code>
     *        object
     * @param panel the <code>PanelPostProcessing</code> which makes the
     *        user interface with post-processing options
     */
    public ThreadPostProcessing(Biclustering biclustering,
                                TreePath selectionPath,
                                PanelPostProcessing panel) {
        this.biclustering = biclustering;
        this.selectionPath = selectionPath;
        postProcessingPanel = panel;
        progressFrame = new ProgressBarFrame(postProcessingPanel.getMainFrame(),
                                             this,
                                             "reading post-processing options...",
                                             true);
    }

    /**
     * Overrides <code>Thread</code>'s <code>run</code> method.
     * Performs the actual work of this thread, that is, applies
     * post-processing algorithms to the group of biclusters in
     * <code>biclustering</code> attribute.
     */
    public void run() {
        if (biclustering == null || selectionPath == null ||
            postProcessingPanel == null) {
            return;
        }

        boolean filter = postProcessingPanel.getFilterCheckBox().isSelected();
        boolean sort = postProcessingPanel.getSortCheckBox().isSelected();

        if (!filter && !sort) {
            return;
        }

        Biclustering bic = null;
        ArrayList<String> infoStrings = new ArrayList<String>();
        ArrayList<String> infoStyles = new ArrayList<String>();

        infoStrings.add("Post-processing:\n");
        infoStyles.add("bold");

        if (filter) {
            int option = postProcessingPanel.getFilterCriteriaList().
                         getSelectedIndex();
            boolean filterOverlapping =
            option == PanelPostProcessing.FILTER_OVERLAPPING_GENES ||
            option == PanelPostProcessing.FILTER_OVERLAPPING_GENES_CONDITIONS;
            int divisor = biclustering.getNumberOfBiclusters();
            TaskPercentages taskPercentages = null;
            if (filterOverlapping) {
                divisor *= biclustering.getNumberOfBiclusters();
                taskPercentages = new TaskPercentages(TaskPercentages.FILTER_OVERLAPPING);
            }
            else {
                taskPercentages = new TaskPercentages(TaskPercentages.FILTER);
            }
            biclustering.setTaskPercentages(taskPercentages);
            biclustering.setProgressBar(progressFrame.getProgressBar());
            biclustering.setSubtaskValues(divisor);

            try {
                switch (option) {
                case PanelPostProcessing.FILTER_TRIVIAL:
                    bic = this.filter_TRIVIAL(biclustering, infoStrings,
                                              infoStyles);
                    break;
                case PanelPostProcessing.FILTER_GENES:
                    bic = this.filterBy_NR_GENES(biclustering, infoStrings,
                                                 infoStyles);
                    break;
                case PanelPostProcessing.FILTER_CONDITIONS:
                    bic = this.filterBy_NR_CONDITIONS(biclustering, infoStrings,
                            infoStyles);
                    break;
                case PanelPostProcessing.FILTER_GENES_CONDITIONS:
                    bic = this.filterBy_NR_GENES_NR_CONDITIONS(biclustering,
                            infoStrings, infoStyles);
                    break;
                case PanelPostProcessing.FILTER_ACV:
                    bic = this.filterBy_ACV_Threshold(biclustering, infoStrings,
                            infoStyles);
                    break;
                case PanelPostProcessing.FILTER_ARV:
                    bic = this.filterBy_ARV_Threshold(biclustering, infoStrings,
                            infoStyles);
                    break;
                case PanelPostProcessing.FILTER_ACV_ARV:
                    bic = this.filterBy_ACV_ARV_Thresholds(biclustering,
                            infoStrings, infoStyles);
                    break;
                case PanelPostProcessing.FILTER_MSR:
                    bic = this.filterBy_MSR_Threshold(biclustering, infoStrings,
                            infoStyles);
                    break;
                case PanelPostProcessing.FILTER_SIZE:
                    bic = this.filterBy_SIZE_Threshold(biclustering,
                            infoStrings,
                            infoStyles);
                    break;
                case PanelPostProcessing.FILTER_SIZE_ARV_ACV:
                    bic = this.filterBy_SIZE_ARV_ACV_Thresholds(biclustering,
                            infoStrings, infoStyles);
                    break;
                case PanelPostProcessing.FILTER_CONDITIONS_GENES_ARV_ACV:
                    bic = this.
                          filterBy_NR_CONDITIONS_NR_GENES_ARV_ACV_Thresholds(
                                  biclustering, infoStrings, infoStyles);
                    break;
                case PanelPostProcessing.FILTER_CONDITIONS_GENES_ARV_ACV_MSR:
                    bic = this.
                          filterBy_NR_CONDITIONS_NR_GENES_ARV_ACV_MSR_Thresholds(
                                  biclustering, infoStrings, infoStyles);
                    break;
                case PanelPostProcessing.FILTER_OVERLAPPING_GENES:
                    bic = this.filterBy_OVERLAPPING_GENES_Threshold(
                            biclustering,
                            infoStrings, infoStyles);
                    break;
                case PanelPostProcessing.FILTER_OVERLAPPING_GENES_CONDITIONS:
                    bic = this.filterBy_OVERLAPPING_GENES_CONDITIONS_Thresholds(
                            biclustering, infoStrings, infoStyles);
                    break;
                case PanelPostProcessing.FILTER_CONSTANT_PATTERN:
                    bic = this.filter_CONSTANT_PATTERN(
                            (BiclusteringInDiscretizedMatrix) biclustering,
                            infoStrings,
                            infoStyles);
                    break;
                case PanelPostProcessing.FILTER_PATTERN_PVALUE_NOT_SIGNIFICANT:
                    bic = this.filter_PATTERN_PVALUE_NOT_SIGNIFICANT(
                            (BiclusteringInDiscretizedMatrix)biclustering,
                            infoStrings, infoStyles);
                    break;
                case PanelPostProcessing.FILTER_PATTERN_COLUMNS_PVALUE_NOT_SIGNIFICANT:
                    bic = this.filter_PATTERN_COLUMNS_PVALUE_NOT_SIGNIFICANT(
                            (BiclusteringInDiscretizedMatrix)biclustering,
                            infoStrings, infoStyles);
                    break;
                }
                biclustering.updatePercentDone();
            } catch (IOException ex) {
                ex.printStackTrace();
                progressFrame.dispose();
                return;
            } catch (Exception ex) {
                ex.printStackTrace();
                progressFrame.dispose();
                postProcessingPanel.displayFilteringErrorMessage(sort);
                return;
            }
            if (bic == null || bic.getSetOfBiclusters().length == 0) {
                progressFrame.dispose();
                JOptionPane.showMessageDialog(postProcessingPanel.getMainFrame(),
                                              "The resulting group has " +
                                              "no biclusters. Therefore, it will" +
                                              " not be added to the dataset tree.",
                                              "Group has no biclusters",
                                              JOptionPane.INFORMATION_MESSAGE);
                return;
            }
        }

        if (sort) {
            if (filter) {
                biclustering = bic;
            }
            TaskPercentages taskPercentages = new TaskPercentages(TaskPercentages.SORT);
            biclustering.setTaskPercentages(taskPercentages);
            biclustering.setProgressBar(progressFrame.getProgressBar());
            this.progressFrame.getProgressBar().setValue(0);

            int option = postProcessingPanel.getSortCriteriaList().
                         getSelectedIndex();
            try {
                switch (option) {
                case PanelPostProcessing.SORT_ACV:
                    bic = this.sortBy_ACV(biclustering, infoStrings, infoStyles);
                    break;
                case PanelPostProcessing.SORT_ARV:
                    bic = this.sortBy_ARV(biclustering, infoStrings, infoStyles);
                    break;
                case PanelPostProcessing.SORT_MSR:
                    bic = this.sortBy_MSR(biclustering, infoStrings, infoStyles);
                    break;
                case PanelPostProcessing.SORT_ARV_ACV_DIFF:
                    bic = this.sortBy_ARV_ACV_DIFF(biclustering, infoStrings,
                            infoStyles);
                    break;
                case PanelPostProcessing.SORT_ARV_MSR_DIFF:
                    bic = this.sortBy_ARV_MSR_DIFF(biclustering, infoStrings,
                            infoStyles);
                    break;
                case PanelPostProcessing.SORT_GENES:
                    bic = this.sortBy_NR_GENES(biclustering, infoStrings,
                                               infoStyles);
                    break;
                case PanelPostProcessing.SORT_CONDITIONS:
                    bic = this.sortBy_NR_CONDITIONS(biclustering, infoStrings,
                            infoStyles);
                    break;
                case PanelPostProcessing.SORT_SIZE:
                    bic = this.sortBy_SIZE(biclustering, infoStrings,
                                           infoStyles);
                    break;
                case PanelPostProcessing.SORT_VARIANCE:
                    bic = this.sortBy_VARIANCE(biclustering, infoStrings,
                                               infoStyles);
                    break;
                case PanelPostProcessing.SORT_BEST_GO_PVALUE:
                    bic = this.sortBy_BEST_GO_PVALUE(biclustering, infoStrings,
                            infoStyles);
                    break;
                case PanelPostProcessing.SORT_BEST_GO_BONFERRONI_PVALUE:
                    bic = this.sortBy_BEST_GO_BONFERRONI_PVALUE(biclustering,
                            infoStrings, infoStyles);
                    break;
                case PanelPostProcessing.SORT_PATTERN_PVALUE:
                    bic = this.sortBy_PATTERN_PVALUE(
                            (BiclusteringInDiscretizedMatrix) biclustering,
                            infoStrings,
                            infoStyles);
                    break;
                case PanelPostProcessing.SORT_PATTERN_COLUMNS_PVALUE:
                    bic = this.sortBy_PATTERN_WITH_COLUMNS_PVALUE(
                            (BiclusteringInDiscretizedMatrix) biclustering,
                            infoStrings, infoStyles);
                    break;
                }
            } catch (IOException ex) {
                progressFrame.dispose();
                return;
            } catch (Exception ex) {
                progressFrame.dispose();
                ex.printStackTrace();
                JOptionPane.showMessageDialog(postProcessingPanel.getMainFrame(),
                                              "Unable to sort biclusters: " +
                                              ex.getMessage() +
                                              ".",
                                              "Unable to sort biclusters",
                                              JOptionPane.ERROR_MESSAGE);
                return;
            }
            if (bic == null || bic.getSetOfBiclusters().length == 0) {
                progressFrame.dispose();
                JOptionPane.showMessageDialog(postProcessingPanel.getMainFrame(),
                                              "The resulting group has " +
                                              "no biclusters. Therefore, it will" +
                                              " not be added to the dataset tree.",
                                              "Group has no biclusters",
                                              JOptionPane.INFORMATION_MESSAGE);
                return;
            }
        }

        progressFrame.setString("copying function analysis graph files...");
        try {
            postProcessingPanel.
                    copyFunctionResultFilesForPostProcessedBiclusterGroup(
                            selectionPath);
            progressFrame.setString(
                    "adding new post-processed group to tree...");
            postProcessingPanel.addPostProcessedBiclusterGroupToTree(bic,
                    infoStrings, infoStyles, selectionPath);
            progressFrame.dispose();
            bic.setTaskPercentages(null);
            bic.setProgressBar(null);
            biclustering.setTaskPercentages(null);
            biclustering.setProgressBar(null);
        }
        catch (Exception ex) {
            progressFrame.dispose();
            JOptionPane.showMessageDialog(postProcessingPanel, "Unable to add the " +
                                          "new post-processed bicluster " +
                                          "group node to the tree.",
                                          "Unable to add bicluster group " +
                                          "node to the tree",
                                          JOptionPane.ERROR_MESSAGE);
            ex.printStackTrace();
        }
    }

    /**
     * Performs trivial biclusters filtering algorithm.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to filter
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new filtered group of biclusters
     * @throws Exception if the group of biclusters cannot be filtered
     */
    public Biclustering filter_TRIVIAL(Biclustering biclustering,
                                       ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        progressFrame.setString("filtering biclusters...");
        bic = PostProcessedBiclustering.new_filter_TRIVIAL(
                biclustering);
        infoStrings.add("   filtered trivial biclusters\n\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs filtering of biclusters with less than a given number
     * of genes. Reads the number of genes from the corresponding input
     * text field located in <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to filter
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new filtered group of biclusters
     * @throws Exception if the group of biclusters cannot be filtered
     */
    public Biclustering filterBy_NR_GENES(Biclustering biclustering,
                                          ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        int genes = 0;
        try {
            genes = postProcessingPanel.readParameterInteger(1,
                    PanelPostProcessing.PARAMETER_GENES);
        } catch (Exception ex) {
            throw new IOException();
        }
        progressFrame.setString("filtering biclusters...");
        bic = PostProcessedBiclustering.new_filter_GENES(
                biclustering, genes);
        infoStrings.add("   filtered biclusters with\n" +
                               "   less than " + genes +
                               " genes\n\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs filtering of biclusters with less than a given number
     * of conditions. Reads the number of conditions from the corresponding
     * input text field located in <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to filter
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new filtered group of biclusters
     * @throws Exception if the group of biclusters cannot be filtered
     */
    public Biclustering filterBy_NR_CONDITIONS(Biclustering biclustering,
                                               ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        int conditions = 0;
        try {
            conditions = postProcessingPanel.readParameterInteger(1,
                    postProcessingPanel.PARAMETER_CONDITIONS);
        } catch (Exception ex) {
            throw new IOException();
        }
        progressFrame.setString("filtering biclusters...");
        bic = PostProcessedBiclustering.new_filter_CONDITIONS(
                biclustering, conditions);
        infoStrings.add("   filtered biclusters with\n" +
                               "   less than " + conditions +
                               " conditions\n\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs filtering of biclusters with less than a given number
     * of genes and a given number of conditions. Reads these numbers
     * from the corresponding input text fields located in
     * <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to filter
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new filtered group of biclusters
     * @throws Exception if the group of biclusters cannot be filtered
     */
    public Biclustering filterBy_NR_GENES_NR_CONDITIONS(
            Biclustering biclustering,
            ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        int genes = 0;
        int conditions = 0;
        try {
            genes = postProcessingPanel.readParameterInteger(1,
                    postProcessingPanel.PARAMETER_GENES);
            conditions = postProcessingPanel.readParameterInteger(1,
                    postProcessingPanel.PARAMETER_CONDITIONS);
        } catch (Exception ex) {
            throw new IOException();
        }

        progressFrame.setString("filtering biclusters...");
        bic = PostProcessedBiclustering.new_filter_GENES(
                biclustering, genes);
        bic = PostProcessedBiclustering.new_filter_CONDITIONS(bic,
                conditions);
        infoStrings.add("   filtered biclusters with\n" +
                               "   less than " + genes + " genes\n" +
                               "   less than " + conditions +
                               " conditions\n\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs filtering of biclusters with an ACV below a given threshold.
     * Reads the threshold from the corresponding input
     * text field located in <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to filter
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new filtered group of biclusters
     * @throws Exception if the group of biclusters cannot be filtered
     */
    public Biclustering filterBy_ACV_Threshold(Biclustering biclustering,
                                               ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        float acv = 0;

        try {
            acv = postProcessingPanel.readParameterFloat(0,
                    postProcessingPanel.PARAMETER_ACV);
        } catch (Exception ex) {
            throw new IOException();
        }

        progressFrame.setString("filtering biclusters...");
        bic = PostProcessedBiclustering.new_filter_ACV(biclustering,
                acv);
        infoStrings.add("   filtered biclusters with\n" +
                               "   average column variance >" + acv +
                               "\n\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs filtering of biclusters with an ARV below a given threshold.
     * Reads the threshold from the corresponding input
     * text field located in <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to filter
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new filtered group of biclusters
     * @throws Exception if the group of biclusters cannot be filtered
     */
    public Biclustering filterBy_ARV_Threshold(Biclustering biclustering,
                                               ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        float arv = 0;
        try {
            arv = postProcessingPanel.readParameterFloat(0,
                    postProcessingPanel.PARAMETER_ARV);
        } catch (Exception ex) {
            throw new IOException();
        }

        progressFrame.setString("filtering biclusters...");
        bic = PostProcessedBiclustering.new_filter_ARV(biclustering,
                arv);
        infoStrings.add("   filtered biclusters with\n" +
                               "   average row variance <" + arv +
                               "\n\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs filtering of biclusters with ACV and ARV below two
     * given thresholds. Reads the thresholds from the corresponding input
     * text fields located in <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to filter
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new filtered group of biclusters
     * @throws Exception if the group of biclusters cannot be filtered
     */
    public Biclustering filterBy_ACV_ARV_Thresholds(Biclustering biclustering,
            ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        float acv = 0;
        float arv = 0;
        try {
            acv = postProcessingPanel.readParameterFloat(0,
                    postProcessingPanel.PARAMETER_ACV);
            arv = postProcessingPanel.readParameterFloat(0,
                    postProcessingPanel.PARAMETER_ARV);
        } catch (Exception ex) {
            throw new IOException();
        }

        progressFrame.setString("filtering biclusters...");
        bic = PostProcessedBiclustering.new_filter_ARV_ACV(
                biclustering, arv, acv);
        infoStrings.add("   filtered biclusters with\n" +
                               "   average column variance >" + acv +
                               "\n" +
                               "   average row variance <" + arv +
                               "\n\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs filtering of biclusters with a MSR below a
     * given threshold. Reads the threshold from the corresponding input
     * text field located in <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to filter
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new filtered group of biclusters
     * @throws Exception if the group of biclusters cannot be filtered
     */
    public Biclustering filterBy_MSR_Threshold(Biclustering biclustering,
                                               ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        float msr = 0;
        try {
            msr = postProcessingPanel.readParameterFloat(0,
                    postProcessingPanel.PARAMETER_MSR);
        } catch (Exception ex) {
            throw new IOException();
        }

        progressFrame.setString("filtering biclusters...");
        bic = PostProcessedBiclustering.new_filter_MSR(biclustering,
                msr);
        infoStrings.add("   filtered biclusters with\n" +
                               "   mean squared residue >" + msr +
                               "\n\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs filtering of biclusters with size below a
     * given threshold. Reads the threshold from the corresponding input
     * text field located in <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to filter
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new filtered group of biclusters
     * @throws Exception if the group of biclusters cannot be filtered
     */
    public Biclustering filterBy_SIZE_Threshold(Biclustering biclustering,
                                                ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        int size = 0;
        try {
            size = postProcessingPanel.readParameterInteger(1,
                    postProcessingPanel.PARAMETER_SIZE);
        } catch (Exception ex) {
            throw new IOException();
        }

        progressFrame.setString("filtering biclusters...");
        bic = PostProcessedBiclustering.new_filter_SIZE(
                biclustering, size);
        infoStrings.add("   filtered biclusters with\n" +
                               "   size <" + size + "\n\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs filtering of biclusters with size, ACV and ARV below three
     * given thresholds. Reads the thresholds from the corresponding input
     * text fields located in <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to filter
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new filtered group of biclusters
     * @throws Exception if the group of biclusters cannot be filtered
     */
    public Biclustering filterBy_SIZE_ARV_ACV_Thresholds(Biclustering
            biclustering,
            ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        int size = 0;
        float arv = 0;
        float acv = 0;

        try {
            size = postProcessingPanel.readParameterInteger(1,
                    postProcessingPanel.PARAMETER_SIZE);
            arv = postProcessingPanel.readParameterFloat(0,
                    postProcessingPanel.PARAMETER_ARV);
            acv = postProcessingPanel.readParameterFloat(0,
                    postProcessingPanel.PARAMETER_ACV);
        } catch (Exception ex) {
            throw new IOException();
        }

        progressFrame.setString("filtering biclusters...");
        bic = PostProcessedBiclustering.new_filter_SIZE_ARV_ACV(
                biclustering, size, arv, acv);
        infoStrings.add("   filtered biclusters with\n" +
                               "   size <" + size + "\n" +
                               "   average column variance >" + acv +
                               "\n" +
                               "   average row variance <" + arv +
                               "\n\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs filtering of biclusters with a number of conditions,
     * a number of genes, ACV and ARV below four
     * given thresholds. Reads the thresholds from the corresponding input
     * text fields located in <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to filter
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new filtered group of biclusters
     * @throws Exception if the group of biclusters cannot be filtered
     */
    public Biclustering filterBy_NR_CONDITIONS_NR_GENES_ARV_ACV_Thresholds(
            Biclustering biclustering,
            ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        int genes = 0, conditions = 0;
        float arv = 0, acv = 0;
        try {
            genes = postProcessingPanel.readParameterInteger(1,
                    postProcessingPanel.PARAMETER_GENES);
            conditions = postProcessingPanel.readParameterInteger(1,
                    postProcessingPanel.PARAMETER_CONDITIONS);
            arv = postProcessingPanel.readParameterFloat(0,
                    postProcessingPanel.PARAMETER_ARV);
            acv = postProcessingPanel.readParameterFloat(0,
                    postProcessingPanel.PARAMETER_ACV);
        } catch (Exception ex) {
            throw new IOException();
        }

        progressFrame.setString("filtering biclusters...");
        bic = PostProcessedBiclustering.
              new_filter_CONDITIONS_GENES_ARV_ACV(biclustering,
                                                  conditions, genes, arv, acv);
        infoStrings.add("   filtered biclusters with\n" +
                               "   less than " + genes + " genes\n" +
                               "   less than " + conditions +
                               " conditions\n" +
                               "   average column variance >" + acv +
                               "\n" +
                               "   average row variance <" + arv +
                               "\n\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs filtering of biclusters with a number of conditions,
     * a number of genes, ACV, ARV and MSR below five
     * given thresholds. Reads the thresholds from the corresponding input
     * text fields located in <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to filter
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new filtered group of biclusters
     * @throws Exception if the group of biclusters cannot be filtered
     */
    public Biclustering filterBy_NR_CONDITIONS_NR_GENES_ARV_ACV_MSR_Thresholds(
            Biclustering biclustering,
            ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        int genes = 0, conditions = 0;
        float arv = 0, acv = 0, msr = 0;
        try {
            genes = postProcessingPanel.readParameterInteger(1,
                    PanelPostProcessing.PARAMETER_GENES);
            conditions = postProcessingPanel.readParameterInteger(1,
                    PanelPostProcessing.PARAMETER_CONDITIONS);
            arv = postProcessingPanel.readParameterFloat(0,
                    PanelPostProcessing.PARAMETER_ARV);
            acv = postProcessingPanel.readParameterFloat(0,
                    PanelPostProcessing.PARAMETER_ACV);
            msr = postProcessingPanel.readParameterFloat(0,
                    PanelPostProcessing.PARAMETER_MSR);
        } catch (Exception ex) {
            throw new IOException();
        }

        progressFrame.setString("filtering biclusters...");
        bic = PostProcessedBiclustering.
              new_filter_CONDITIONS_GENES_ARV_ACV_MSR(biclustering,
                conditions, genes, arv, acv, msr);
        infoStrings.add("   filtered biclusters with\n" +
                               "   less than " + genes + " genes\n" +
                               "   less than " + conditions +
                               " conditions\n" +
                               "   average column variance >" + acv +
                               "\n"
                               + "   average row variance <" + arv +
                               "\n" +
                               "   mean squared residue >" + msr +
                               "\n\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs filtering of biclusters with a % of overlapping genes below a
     * given threshold. Reads the threshold from the corresponding input
     * text field located in <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to filter
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new filtered group of biclusters
     * @throws Exception if the group of biclusters cannot be filtered
     */
    public Biclustering filterBy_OVERLAPPING_GENES_Threshold(Biclustering
            biclustering,
            ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        int percent = 1;
        try {
            percent = postProcessingPanel.readParameterInteger(1,
                    100, postProcessingPanel.PARAMETER_SIMILARITY);
        } catch (Exception ex) {
            throw new IOException();
        }

        progressFrame.setString("filtering biclusters...");
        bic = PostProcessedBiclustering.new_filter_OVERLAPPING_ROWS(
                biclustering, (float) ((percent * 1.0) / 100.0));
        infoStrings.add("   filtered biclusters with\n" +
                               "   row similarity > " + percent +
                               "%\n\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs filtering of biclusters with a % of overlapping genes
     * and a % of overlapping conditions below two given thresholds.
     * Reads the thresholds from the corresponding input
     * text fields located in <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to filter
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new filtered group of biclusters
     * @throws Exception if the group of biclusters cannot be filtered
     */
    public Biclustering filterBy_OVERLAPPING_GENES_CONDITIONS_Thresholds(
            Biclustering biclustering,
            ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        int percent = 1;
        try {
            percent = postProcessingPanel.readParameterInteger(1,
                    100, postProcessingPanel.PARAMETER_SIMILARITY);
        } catch (Exception ex) {
            throw new IOException();
        }

        progressFrame.setString("filtering biclusters...");
        bic = PostProcessedBiclustering.
              new_filter_OVERLAPPING_ROWS_COLUMNS(biclustering,
                                                  (float) ((percent * 1.0) /
                100.0));
        infoStrings.add("   filtered biclusters with\n" +
                               "   row and column similarity > " +
                               percent + "%\n\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs filtering of biclusters with a constant pattern
     * and a % of overlapping conditions below two given thresholds.
     * Reads the thresholds from the corresponding input
     * text fields located in <code>postProcessingPanel</code>.
     * This method is suited only for biclusters extracted from
     * discretized epression matrices.
     *
     * @param biclustering <code>BiclusteringInDiscretizedMatrix</code>
     *        the group of biclusters to filter
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new filtered group of biclusters
     * @throws Exception if the group of biclusters cannot be filtered
     */
    public Biclustering filter_CONSTANT_PATTERN(BiclusteringInDiscretizedMatrix
                                                biclustering,
                                                ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        progressFrame.setString("filtering biclusters...");
        bic = PostProcessedBiclusteringInDiscretizedMatrix.
              new_filter_CONSTANT_EXPRESSION_PATTERN(biclustering);
        infoStrings.add("   filtered biclusters with\n" +
                               "   constant expression pattern\n\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Filters the biclusters of the given biclusters' group
     * (<code>BiclusteringInDiscretizedMatrix</code>) for which
     * the pattern p-value is not significant.
     *
     * @param biclustering <code>BiclusteringInDiscretizedMatrix</code>
     *             (the biclusters' group)
     * @param infoStrings <code>ArrayList</code> information strings
     * @param infoStyles <code>ArrayList</code> styles for information strings
     * @return Biclustering <code>biclustering</code> object after filtering
     * @throws Exception
     */
    public Biclustering filter_PATTERN_PVALUE_NOT_SIGNIFICANT(
            BiclusteringInDiscretizedMatrix biclustering,
            ArrayList<String> infoStrings, ArrayList<String> infoStyles)
            throws Exception {
        Biclustering bic = null;
        progressFrame.setString("filtering biclusters...");

        int markovChainOrder = 1;
        float pValueThreshold = (float)0.01;
        try {
            markovChainOrder = postProcessingPanel.readParameterInteger(
                    1,
                    ((DiscretizedExpressionMatrix)
                     biclustering.getMatrix()).
                    getNumberOfConditions(),
                    postProcessingPanel.PARAMETER_MARKOV_FILTER);
            pValueThreshold = postProcessingPanel.readParameterFloat(0,
                    1, postProcessingPanel.PARAMETER_PVALUE);

        } catch (Exception ex) {
            throw new IOException();
        }

        bic = PostProcessedBiclusteringInDiscretizedMatrix.
              new_filter_BiclusterPattern_BonferroniCorrected_pValue(
                  biclustering, pValueThreshold, markovChainOrder);
        infoStrings.add("   filtered biclusters without\n" +
                        "   significant pattern\n\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Filters the biclusters of the given biclusters' group
     * (<code>BiclusteringInDiscretizedMatrix</code>) for which
     * the pattern p-value with columns is not significant.
     *
     * @param biclustering <code>BiclusteringInDiscretizedMatrix</code>
     *             (the biclusters' group)
     * @param infoStrings <code>ArrayList</code> information strings
     * @param infoStyles <code>ArrayList</code> styles for information strings
     * @return Biclustering <code>biclustering</code> object after filtering
     * @throws Exception
     */
    public Biclustering filter_PATTERN_COLUMNS_PVALUE_NOT_SIGNIFICANT(
            BiclusteringInDiscretizedMatrix biclustering,
            ArrayList<String> infoStrings, ArrayList<String> infoStyles)
            throws Exception {
        Biclustering bic = null;
        progressFrame.setString("filtering biclusters...");

        int markovChainOrder = 1;
        float pValueThreshold = (float)0.01;
        try {
            markovChainOrder = postProcessingPanel.readParameterInteger(
                    1,
                    ((DiscretizedExpressionMatrix)
                     biclustering.getMatrix()).
                    getNumberOfConditions(),
                    postProcessingPanel.PARAMETER_MARKOV_FILTER);
            pValueThreshold = postProcessingPanel.readParameterFloat(0,
                    1, postProcessingPanel.PARAMETER_PVALUE);

        } catch (Exception ex) {
            throw new IOException();
        }

        bic = PostProcessedBiclusteringInDiscretizedMatrix.
              new_filter_BiclusterPatternWithColumns_BonferroniCorrected_pValue(
                  biclustering, pValueThreshold, markovChainOrder);
        infoStrings.add("   filtered biclusters without\n" +
                        "   significant pattern with columns\n\n");
        infoStyles.add("regular");
        return bic;
    }


    /**
     * Performs sorting of biclusters by their ACV. Checks which
     * order (ascending or descending) should be applied by accessing
     * the selected property of ascending and descending radio buttons
     * located in <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to sort
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new sorted group of biclusters
     * @throws Exception if the group of biclusters cannot be sorted
     */
    public Biclustering sortBy_ACV(Biclustering biclustering,
                                   ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        if (postProcessingPanel.getAscendingButton().isSelected()) {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_ACV_ASC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   average column variance (ascending)\n");
            infoStyles.add("regular");
        } else {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_ACV_DESC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   average column variance (descending)\n");
            infoStyles.add("regular");
        }
        return bic;
    }

    /**
     * Performs sorting of biclusters by their ARV. Checks which
     * order (ascending or descending) should be applied by accessing
     * the selected property of ascending and descending radio buttons
     * located in <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to sort
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new sorted group of biclusters
     * @throws Exception if the group of biclusters cannot be sorted
     */
    public Biclustering sortBy_ARV(Biclustering biclustering,
                                   ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        if (postProcessingPanel.getAscendingButton().isSelected()) {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_ARV_ASC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   average row variance (ascending)\n");
            infoStyles.add("regular");
        } else {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_ARV_DESC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   average row variance (descending)\n");
            infoStyles.add("regular");
        }
        return bic;
    }

    /**
     * Performs sorting of biclusters by their MSR. Checks which
     * order (ascending or descending) should be applied by accessing
     * the selected property of ascending and descending radio buttons
     * located in <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to sort
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new sorted group of biclusters
     * @throws Exception if the group of biclusters cannot be sorted
     */
    public Biclustering sortBy_MSR(Biclustering biclustering,
                                   ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        if (postProcessingPanel.getAscendingButton().isSelected()) {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_MSR_ASC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   mean squared residue (ascending)\n");
            infoStyles.add("regular");
        } else {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_MSR_DESC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   mean squared residue (descending)\n");
            infoStyles.add("regular");
        }
        return bic;
    }

    /**
     * Performs sorting of biclusters by the difference of their ARV
     * and ACV (ARV-ACV). Checks which order (ascending or descending) should
     * be applied by accessing the selected property of ascending
     * and descending radio buttons located in
     * <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to sort
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new sorted group of biclusters
     * @throws Exception if the group of biclusters cannot be sorted
     */
    public Biclustering sortBy_ARV_ACV_DIFF(Biclustering biclustering,
                                            ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        if (postProcessingPanel.getAscendingButton().isSelected()) {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_ARV_ACV_ASC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   (ARV - ACV) difference (ascending)\n");
            infoStyles.add("regular");
        } else {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_ARV_ACV_DESC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   (ARV - ACV) difference (descending)\n");
            infoStyles.add("regular");
        }
        return bic;
    }

    /**
     * Performs sorting of biclusters by the difference of their ARV
     * and MSR (ARV-MSR). Checks which order (ascending or descending) should
     * be applied by accessing the selected property of ascending
     * and descending radio buttons located in
     * <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to sort
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new sorted group of biclusters
     * @throws Exception if the group of biclusters cannot be sorted
     */
    public Biclustering sortBy_ARV_MSR_DIFF(Biclustering biclustering,
                                            ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        if (postProcessingPanel.getAscendingButton().isSelected()) {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_ARV_MSR_ASC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   (ARV - MSR) difference (ascending)\n");
            infoStyles.add("regular");
        } else {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_ARV_MSR_DESC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   (ARV - MSR) difference (descending)\n");
            infoStyles.add("regular");
        }
        return bic;
    }

    /**
     * Performs sorting of biclusters by their number of genes.
     * Checks which order (ascending or descending) should
     * be applied by accessing the selected property of ascending
     * and descending radio buttons located in
     * <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to sort
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new sorted group of biclusters
     * @throws Exception if the group of biclusters cannot be sorted
     */
    public Biclustering sortBy_NR_GENES(Biclustering biclustering,
                                        ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        if (postProcessingPanel.getAscendingButton().isSelected()) {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_GENES_ASC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   number of genes (ascending)\n");
            infoStyles.add("regular");
        } else {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_GENES_DESC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   number of genes (descending)\n");
            infoStyles.add("regular");
        }
        return bic;
    }

    /**
     * Performs sorting of biclusters by their number of conditions.
     * Checks which order (ascending or descending) should
     * be applied by accessing the selected property of ascending
     * and descending radio buttons located in
     * <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to sort
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new sorted group of biclusters
     * @throws Exception if the group of biclusters cannot be sorted
     */
    public Biclustering sortBy_NR_CONDITIONS(Biclustering biclustering,
                                             ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        if (postProcessingPanel.getAscendingButton().isSelected()) {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_CONDITIONS_ASC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   number of time points (ascending)\n");
            infoStyles.add("regular");
        } else {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.
                  new_sort_CONDITIONS_DESC(
                          biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   number of time points (descending)\n");
            infoStyles.add("regular");
        }
        return bic;
    }

    /**
     * Performs sorting of biclusters by their sizes.
     * Checks which order (ascending or descending) should
     * be applied by accessing the selected property of ascending
     * and descending radio buttons located in
     * <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to sort
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new sorted group of biclusters
     * @throws Exception if the group of biclusters cannot be sorted
     */
    public Biclustering sortBy_SIZE(Biclustering biclustering,
                                    ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        if (postProcessingPanel.getAscendingButton().isSelected()) {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_SIZE_ASC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   size (ascending)\n");
            infoStyles.add("regular");
        } else {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_SIZE_DESC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   size (descending)\n");
            infoStyles.add("regular");
        }
        return bic;
    }

    /**
     * Performs sorting of biclusters by their variances.
     * Checks which order (ascending or descending) should
     * be applied by accessing the selected property of ascending
     * and descending radio buttons located in
     * <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to sort
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new sorted group of biclusters
     * @throws Exception if the group of biclusters cannot be sorted
     */
    public Biclustering sortBy_VARIANCE(Biclustering biclustering,
                                        ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        if (postProcessingPanel.getAscendingButton().isSelected()) {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_VAR_ASC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   variance (ascending)\n");
            infoStyles.add("regular");
        } else {
            progressFrame.setString("sorting biclusters...");
            bic = PostProcessedBiclustering.new_sort_VAR_DESC(
                    biclustering);
            infoStrings.add(
                    "   biclusters sorted by\n" +
                    "   variance (descending)\n");
            infoStyles.add("regular");
        }
        return bic;
    }

    /**
     * Performs sorting of biclusters by their best GO p-values
     * in ascending order.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to sort
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new sorted group of biclusters
     * @throws Exception if the group of biclusters cannot be sorted
     */
    public Biclustering sortBy_BEST_GO_PVALUE(Biclustering biclustering,
                                              ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        progressFrame.setString("sorting biclusters...");
        bic = PostProcessedBiclustering.new_sort_Best_GO_pValue_ASC(
            biclustering, postProcessingPanel.getMainFrame().
            getPanelAnalyzingFunctionTable().
            getGOFrontEndForSelectionPath(selectionPath));
        infoStrings.add(
                "   biclusters sorted by\n" +
                "   best p-value (ascending)\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs sorting of biclusters by their best GO Bonferroni
     * corrected p-values in ascending order.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to sort
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new sorted group of biclusters
     * @throws Exception if the group of biclusters cannot be sorted
     */
    public Biclustering sortBy_BEST_GO_BONFERRONI_PVALUE(Biclustering
            biclustering,
            ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        progressFrame.setString("sorting biclusters...");
        bic = PostProcessedBiclustering.
              new_sort_Best_GO_BonferroniCorrected_pValue_ASC(
                  biclustering, postProcessingPanel.getMainFrame().
                  getPanelAnalyzingFunctionTable().
                  getGOFrontEndForSelectionPath(selectionPath));
        infoStrings.add(
                "   biclusters sorted by\n" +
                "   best Bonferroni corrected p-value (ascending)\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs sorting of biclusters by their pattern p-value
     * in ascending order. Retrieves the Markov chain order
     * from the corresponding input text field located in
     * <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to sort
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new sorted group of biclusters
     * @throws Exception if the group of biclusters cannot be sorted
     */
    public Biclustering sortBy_PATTERN_PVALUE(
            BiclusteringInDiscretizedMatrix biclustering,
            ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        int markovChainOrder = 1;
        try {
            markovChainOrder = postProcessingPanel.
                               readParameterInteger(1,
                    ((DiscretizedExpressionMatrix)
                     biclustering.getMatrix()).
                    getNumberOfConditions(),
                    postProcessingPanel.PARAMETER_MARKOV_SORT);
        } catch (Exception ex) {
            throw new IOException();
        }

        progressFrame.setString("sorting biclusters...");
        bic = PostProcessedBiclusteringInDiscretizedMatrix.
              new_sort_BiclusterPattern_pValue_ASC(
                      biclustering, markovChainOrder);
        infoStrings.add(
                "   biclusters sorted by\n" +
                "   bicluster pattern p-value (ascending)\n");
        infoStyles.add("regular");
        return bic;
    }

    /**
     * Performs sorting of biclusters by their pattern with
     * columns p-value in ascending order. Retrieves the
     * Markov chain order from the corresponding input text
     * field located in <code>postProcessingPanel</code>.
     *
     * @param biclustering <code>Biclustering</code> the group of
     *        biclusters to sort
     * @param infoStrings <code>ArrayList</code> the set of information
     *        strings for the group of biclusters
     * @param infoStyles <codeArrayList</code> the set of styles for the
     *        information strings of the biclusters' group
     * @return Biclustering the new sorted group of biclusters
     * @throws Exception if the group of biclusters cannot be sorted
     */
    public Biclustering sortBy_PATTERN_WITH_COLUMNS_PVALUE(
            BiclusteringInDiscretizedMatrix biclustering,
            ArrayList<String> infoStrings,
            ArrayList<String> infoStyles) throws Exception {
        Biclustering bic = null;
        int markovChainOrder = 1;
        try {
            markovChainOrder = postProcessingPanel.readParameterInteger(
                    1,
                    ((DiscretizedExpressionMatrix)
                     biclustering.getMatrix()).
                    getNumberOfConditions(),
                    postProcessingPanel.PARAMETER_MARKOV_SORT);
        } catch (Exception ex) {
            throw new IOException();
        }

        progressFrame.setString("sorting biclusters...");
        biclustering.setProgressBar(progressFrame.getProgressBar());
        bic = PostProcessedBiclusteringInDiscretizedMatrix.
              new_sort_BiclusterPatternWithColumns_pValue_ASC(
                      biclustering,
                      markovChainOrder);
        infoStrings.add(
                "   biclusters sorted by\n" +
                "   bicluster pattern with columns p-value (ascending)\n");
        infoStyles.add("regular");
        return bic;
    }
}
